/** @file */

#ifndef _DYNAMIXEL_COMMON_H
#define _DYNAMIXEL_COMMON_H

#define     MAX_DATA_LENGTH     255

#define     MAX_HEADER_LENGTH   6

// errors
#define     DYN_SUCCESS         0
#define     DYN_BAD_FORMAT      1
#define     DYN_NO_DEVICE       2
#define     DYN_TIMEOUT         3
#define     DYN_COMM_ERROR      4
#define     DYN_BUSY            5

/**
 * \brief Possible return levels
 *
 */
typedef enum{no_return=0x00,/*!< Only the PING command will return an aswer */
             return_only_read=0x01,/*!< Only the read commands will return an answer */
             return_all=0x02/*!< All commands will return an aswer. This is the default value */
            } return_level_t;

// possible packet types
/**
 * \brief Possible instructions
 *
 */
typedef enum{DYN_PING=0x01,/*!< Ping. Instruction to detect the presence of dynamixel devices */
             DYN_READ=0x02,/*!< Read. Instruction to read a consecutive number of registers starting from a given address from a single device */
             DYN_WRITE=0x03,/*!< Write. Instruction to write a consecutive number of registers starting at a given address to a given device*/
             DYN_REG_WRITE=0x04,/*!< Registerd write. Instruction to execute a delayed write operation */
             DYN_ACTION=0x05,/*!< Action. Instruction to actually load the data of a previous registeres write instruction*/
             DYN_RESET=0x06,/*!< Reset. Instruction to reset the device to a default state */
             DYN_SYNC_READ=0x82,/*!< Synchronous read. Instruction to read a consecutive number of registers starting from a given address from multiples devices */
             DYN_SYNC_WRITE=0x83,/*!< Synchronous write. Instruction to write a consecutive number of registers starting at a given address to multiple devices */
             DYN_BULK_READ=0x92,/*!< Bulk read. Instruction to read a set of consecutive registers from multiples devices, starting at a different address for each device */
             DYN_BULK_WRITE=0x93/*!< Bulk write. Instruction to write a set of consecutive registers to multiples devices, starting at a different address for each device */
            } TDynInstruction;

// boradcast ID
#define     DYN_BROADCAST_ID    0xFE/*!< Broadcast ID */

// status packet
/**
 * \brief Possible errors
 *
 */
typedef enum{DYN_NO_ERROR=0x00,/*!< No error */
             DYN_INST_ERROR=0x40,/*!< Invalid Instruction identifier */
             DYN_OVERLOAD_ERROR=0x20,/*!< Servo exceeded the maximum torque value */
             DYN_CHECKSUM_ERROR=0x10,/*!< Invalid checksum on the instruction packet */
             DYN_RANGE_ERROR=0x08,/*!< Register value out of the allowed range */
             DYN_OVERTEMP_ERROR=0x04,/*!< Temperature exceeded the maximum value */
             DYN_ANGLE_ERROR=0x02,/*!< Target position out of range */
             DYN_VOLTAGE_ERROR=0x01/*!< Power supply voltage out of range */
            } TDynError;

/**
 * \brief Dynamixel versions
 *
 */
typedef enum{DYN_VER1=0x01,/*!< Dynamixel version 1.0 */
             DYN_VER2=0x02/*!< Dynamixel version 2.0 */
            } TDynVersion;
#endif
