#include "dynamixel_master.h"

/* private functions */
unsigned char dyn_master_irq_receive_cb(TDynamixelMaster *dyn_master,unsigned char byte)
{
  if(dyn_master!=0x00000000)
  {
    if(dyn_master->comm_dev->time!=0x00000000)
      time_set_timeout(dyn_master->comm_dev->time,dyn_master->rx_timeout_ms*1000);
    switch(dyn_master->received_bytes)
    {
      case 0: if(byte==0xFF)
              {
                dyn_master->rx_buffer[dyn_master->received_bytes]=byte;
                dyn_master->received_bytes++;
              }
              break;
      case 1: if(byte==0xFF)
              {
                dyn_master->rx_buffer[dyn_master->received_bytes]=byte;
                dyn_master->received_bytes++;
              }
              else
                dyn_master->received_bytes--;
              break;
      case 2: if(byte==0xFD)// version 2 header
              {
                if(dyn_master->version==DYN_VER2)// the module is configured for version 2
                {
                  dyn_master->rx_buffer[dyn_master->received_bytes]=byte;
                  dyn_master->received_bytes++;
                }
                else
                  dyn_master->received_bytes=0;// ignore packet and restart synchronization
              }
              else if(byte!=0xFF)
              {
                dyn_master->rx_buffer[dyn_master->received_bytes]=byte;
                dyn_master->received_bytes++;
              }
              break;
      case 3: dyn_master->rx_buffer[dyn_master->received_bytes]=byte;
              if(dyn_master->version==DYN_VER1)
              {
                dyn_master->op_length=byte;
                dyn_master->received_bytes=0;
                /* finish reception by IRQ */
                comm_cancel_irq_receive(dyn_master->comm_dev);
                /* enable dma RX */
                comm_receive_dma(dyn_master->comm_dev,&dyn_master->rx_buffer[4],dyn_master->op_length);
              }
              else
                dyn_master->received_bytes++;
              break;
      case 4: dyn_master->rx_buffer[dyn_master->received_bytes]=byte;
              dyn_master->received_bytes++;
              break;
      case 5: dyn_master->rx_buffer[dyn_master->received_bytes]=byte;
              dyn_master->received_bytes++;
              dyn_master->op_length=byte;
              break;
      case 6: dyn_master->rx_buffer[dyn_master->received_bytes]=byte;
              dyn_master->received_bytes++;
              dyn_master->op_length+=(byte<<8);
              dyn_master->received_bytes=0;
              /* finish reception by IRQ */
              comm_cancel_irq_receive(dyn_master->comm_dev);
              /* enable dma RX */
              comm_receive_dma(dyn_master->comm_dev,&dyn_master->rx_buffer[7],dyn_master->op_length);
              break;
      default: break;
    }
  }

  return 0x00;
}

unsigned char dyn_master_dma_send_cb(TDynamixelMaster *dyn_master)
{
  if(dyn_master!=0x00000000)
  {
    dyn_master->set_rx_mode();
    if(dyn_master->rx_no_answer)
      dyn_master->rx_no_answer=0x00;
    else
    {
      /* enable RX interrupts */
      comm_receive_irq(dyn_master->comm_dev,0);
    }
  }

  return 0x00;
}

unsigned char dyn_master_dma_receive_cb(TDynamixelMaster *dyn_master)
{
  if(dyn_master!=0x00000000)
  {
    dyn_master->rx_num_packets--;
    if(dyn_master->rx_num_packets==0x00)
    {
      if(dyn_master->comm_dev->time!=0x00000000)
        time_cancel_timeout(dyn_master->comm_dev->time);
      dyn_master->packet_ready=0x01;
    }
    else
    {
      /* enable RX interrupts */
      comm_receive_irq(dyn_master->comm_dev,0);
      dyn_master->packet_ready=0x01;
    }
  }

  return 0x00;
}

void dummy_dyn_master_set_tx_mode(void)
{

}

void dummy_dyn_master_set_rx_mode(void)
{

}

unsigned char dyn_master_wait_transmission(TDynamixelMaster *master)
{
  unsigned char error;

  if(master!=0x00000000)
  {
    while((error=comm_is_send_done(master->comm_dev))==COMM_BUSY);
  
    if(error==COMM_SUCCESS)
      return DYN_SUCCESS;
    else
      return DYN_COMM_ERROR;
  }
  
  else 
    return DYN_COMM_ERROR;
}

unsigned char dyn_master_is_transmission_done(TDynamixelMaster *master)
{
  unsigned char error;

  if(master!=0x00000000)
  {
    error=comm_is_send_done(master->comm_dev);
    if(error==COMM_BUSY)
      return DYN_BUSY;
    else if(error==COMM_SUCCESS)
    {
      if(master->comm_dev->time!=0x00000000)
      {
        // start the new timeout
        time_set_timeout(master->comm_dev->time,master->rx_timeout_ms*1000);
      }
      return DYN_SUCCESS;
    }
    else
    {
      master->set_rx_mode();
      return DYN_COMM_ERROR;
    }
  }
  else
    return DYN_COMM_ERROR;
}

unsigned char dyn_master_send(TDynamixelMaster *master)
{
  unsigned char error;

  if(master!=0x00000000)
  {
    // wait until any previous transmission ends
    if((error=dyn_master_wait_transmission(master))!=DYN_SUCCESS)
      return error;
    // set the DMA transfer
    if(master->version==DYN_VER1)
    {
      comm_send_dma(master->comm_dev,master->tx_buffer,dyn_get_length(master->tx_buffer)+4);
      return DYN_SUCCESS;
    }
    else
    {
      comm_send_dma(master->comm_dev,master->tx_buffer,dyn2_get_length(master->tx_buffer)+7);
      return DYN_SUCCESS;
    }
  }
  else
    return DYN_COMM_ERROR;
} 

unsigned char dyn_master_wait_reception(TDynamixelMaster *master)
{
  if(master!=0x00000000)
  {
    if(master->comm_dev->time!=0x00000000)
    {
      // start the new timeout
      time_set_timeout(master->comm_dev->time,master->rx_timeout_ms*1000);
    }
    // wait for the status packet
    while(!master->packet_ready)
    {
      if(master->comm_dev->time!=0x00000000)
      {
        if(time_is_timeout(master->comm_dev->time))
        {
          comm_cancel_irq_receive(master->comm_dev);
          comm_cancel_dma_receive(master->comm_dev);
          master->received_bytes=0x00;
          return DYN_TIMEOUT;
        }
      }
    }
    master->packet_ready=0x00;
    // check the input packet checksum
    if(master->version==DYN_VER1)
    {
      if(dyn_check_checksum(master->rx_buffer)==0xFF)
        return dyn_get_status_error(master->rx_buffer);
      else
        return DYN_CHECKSUM_ERROR;
    }
    else
    {
      if(dyn2_check_checksum(master->rx_buffer)==0x01)
        return dyn2_get_status_error(master->rx_buffer);
      else
        return DYN_CHECKSUM_ERROR;
    }
  }
  else
    return DYN_COMM_ERROR;
}

unsigned char dyn_master_is_reception_done(TDynamixelMaster *master)
{
  if(master!=0x00000000)
  {
    if(!master->packet_ready)
    {
      if(master->comm_dev->time!=0x00000000)
      {
        if(time_is_timeout(master->comm_dev->time))
        {
          comm_cancel_irq_receive(master->comm_dev);
          comm_cancel_dma_receive(master->comm_dev);
          master->received_bytes=0x00;
          return DYN_TIMEOUT;
        }
        else
          return DYN_BUSY;
      }
      else
        return DYN_BUSY;
    }
    else
    {
      master->packet_ready=0x00;
      return DYN_SUCCESS;
    }
  }
  else
    return DYN_COMM_ERROR;
}

unsigned char dyn_master_start_read_table(TDynamixelMaster *master,unsigned char id,unsigned short int address,unsigned short int length,unsigned char *data)
{
  unsigned char error=DYN_COMM_ERROR;

  if(master!=0x00000000)
  {
    // generate the read packet for the desired device
    if(master->version==DYN_VER1)
      dyn_init_read_packet(master->tx_buffer,id,address,length);
    else
      dyn2_init_read_packet(master->tx_buffer,id,address,length);
    master->rx_num_packets=0x01;
    if(master->return_level==no_return || id==DYN_BROADCAST_ID)
      master->rx_no_answer=0x01;
    else
      master->rx_no_answer=0x00;
    // enable transmission
    master->set_tx_mode();
    // send the data
    if((error=dyn_master_send(master))!=DYN_SUCCESS)
      master->set_rx_mode();
  } 

  return error;
}

unsigned char dyn_master_is_read_table_done(TDynamixelMaster *master,unsigned char id,unsigned short int address,unsigned short int length,unsigned char *data)
{
  unsigned char error=DYN_COMM_ERROR;

  if(master!=0x00000000)
  {
    if((error=dyn_master_is_transmission_done(master))==DYN_SUCCESS)
    {
      // wait for the replay within the given timeout
      if(master->return_level!=no_return && id!=DYN_BROADCAST_ID)
      {
        if((error=dyn_master_is_reception_done(master))==DYN_SUCCESS)
        {
          // check the input packet checksum
          if(master->version==DYN_VER1)
          {
            if(dyn_check_checksum(master->rx_buffer)==0xFF)
            {
              if((error=dyn_get_status_error(master->rx_buffer))==DYN_SUCCESS)
              {
                if(dyn_get_read_status_data(master->rx_buffer,data)!=length)// not enough data
                  return DYN_INST_ERROR;
                else
                  return DYN_SUCCESS;
              }
              else 
                return error;
            }
            else
              return DYN_CHECKSUM_ERROR;
          }
          else
          {
            if(dyn2_check_checksum(master->rx_buffer)==0x01)
            {
              if((error=dyn2_get_status_error(master->rx_buffer))==DYN_SUCCESS)
              {
                if(dyn2_get_read_status_data(master->rx_buffer,data)!=length)// not enough data
                  return DYN_INST_ERROR;
                else
                  return DYN_SUCCESS;
              }
              else
                return error;
            }
            else
              return DYN_CHECKSUM_ERROR;
          }
        }
        else
          return error;
      }
      else
        return DYN_SUCCESS;
    }
    else
      return error;
  }
  else
    return error;
}

unsigned char dyn_master_start_write_table(TDynamixelMaster *master,unsigned char id,unsigned short int address,unsigned short int length,unsigned char *data)
{
  unsigned char error=DYN_COMM_ERROR;

  if(master!=0x00000000)
  {
    // generate the read packet for the desired device
    if(master->version==DYN_VER1)
      dyn_init_write_packet(master->tx_buffer,id,address,length,data);
    else
      dyn2_init_write_packet(master->tx_buffer,id,address,length,data);
    master->rx_num_packets=0x01;
    if(master->return_level==return_all && id!=DYN_BROADCAST_ID)
      master->rx_no_answer=0x00;
    else
      master->rx_no_answer=0x01;
    // enable transmission
    master->set_tx_mode();
    // send the data
    if((error=dyn_master_send(master))!=DYN_SUCCESS)
      master->set_rx_mode();
  }

  return error;
}

unsigned char dyn_master_is_write_table_done(TDynamixelMaster *master,unsigned char id,unsigned short int address,unsigned short int length,unsigned char *data)
{
  unsigned char error=DYN_COMM_ERROR;

  if(master!=0x00000000)
  {
    if((error=dyn_master_is_transmission_done(master))==DYN_SUCCESS)
    {
      // wait for the replay within the given timeout
      if(master->return_level==return_all && id!=DYN_BROADCAST_ID)
      {
        if((error=dyn_master_is_reception_done(master))==DYN_SUCCESS)
        {
          // check the input packet checksum
          if(master->version==DYN_VER1)
          {
            if(dyn_check_checksum(master->rx_buffer)==0xFF)
              return dyn_get_status_error(master->rx_buffer);
            else
              return DYN_CHECKSUM_ERROR;
          }
          else
          {
            if(dyn2_check_checksum(master->rx_buffer)==0x01)
              return dyn2_get_status_error(master->rx_buffer);
            else
              return DYN_CHECKSUM_ERROR;
          }
        }
        else
          return error;
      }
      else
        return DYN_SUCCESS;
    }
    else
      return error;
  }
  else
    return error;
}

unsigned char dyn_master_start_sync_write(TDynamixelMaster *master,unsigned char num,unsigned char *ids,unsigned short int address, unsigned short int length, unsigned char * const data[])
{
  unsigned char error=DYN_COMM_ERROR;

  if(master!=0x00000000) 
  {
    // generate the read packet for the desired device
    if(master->version==DYN_VER1)
      dyn_init_sync_write_packet(master->tx_buffer,num,ids,address,length,data);
    else
      dyn2_init_sync_write_packet(master->tx_buffer,num,ids,address,length,data);
    master->rx_num_packets=0x01;
    master->rx_no_answer=0x00;
    // enable transmission
    master->set_tx_mode();
    // send the data
    if((error=dyn_master_send(master))!=DYN_SUCCESS)
      master->set_rx_mode();
  }

  return error;
}

unsigned char dyn_master_is_sync_write_done(TDynamixelMaster *master,unsigned char num,unsigned char *ids,unsigned short int address, unsigned short int length,unsigned char * const data[])
{
  unsigned char error=DYN_COMM_ERROR;

  if(master!=0x00000000)
  {
    error=comm_is_send_done(master->comm_dev);
    if(error==COMM_BUSY)
      return DYN_BUSY;
    else if(error==COMM_SUCCESS)
      return DYN_SUCCESS;
    else
    {
      master->set_rx_mode();
      return DYN_COMM_ERROR;
    }
  }
  else
    return error;
}

unsigned char dyn_master_start_sync_read(TDynamixelMaster *master,unsigned char num,unsigned char *ids,unsigned short int address, unsigned short int length, unsigned char * const data[])
{
  unsigned char error=DYN_COMM_ERROR;

  if(master!=0x00000000)
  {
    // generate the read packet for the desired device
    if(master->version==DYN_VER2)
    { 
      dyn2_init_sync_read_packet(master->tx_buffer,num,ids,address,length);
      master->rx_num_packets=num;
      if(master->return_level==no_return)
        master->rx_no_answer=0x01;
      else
        master->rx_no_answer=0x00;
      // enable transmission
      master->set_tx_mode();
      // send the data
      if((error=dyn_master_send(master))!=DYN_SUCCESS)
        master->set_rx_mode();

      return error;
    }
    else
      return DYN_INST_ERROR;
  }
  else
    return error;
}

unsigned char dyn_master_is_sync_read_done(TDynamixelMaster *master,unsigned char num,unsigned char *ids,unsigned short int address, unsigned short int length, unsigned char * const data[])
{
  static unsigned char num_dev_done=0;
  unsigned char error=DYN_COMM_ERROR;

  if(master!=0x00000000)
  {
    // generate the read packet for the desired device
    if(master->version==DYN_VER2)
    {
      if((error=dyn_master_is_transmission_done(master))==DYN_SUCCESS)
      {
        // wait for the replay within the given timeout
        if(master->return_level!=no_return)
        {
          if(num_dev_done<num)
          {
            if((error=dyn_master_is_reception_done(master))==DYN_SUCCESS)
            {
              if(dyn2_check_checksum(master->rx_buffer)==0x01)
              {
                if((error=dyn2_get_status_error(master->rx_buffer))==DYN_SUCCESS)
                {
                  while(dyn2_get_read_status_id(master->rx_buffer)!=ids[num_dev_done]) num_dev_done++;
                  if(dyn2_get_read_status_data(master->rx_buffer,(unsigned char *)data[num_dev_done])!=length)// not enough data
                    return DYN_INST_ERROR;
                  else
                  {
                    num_dev_done++;
                    return DYN_BUSY;
                  }
                }
                else
                  return error;
              }
              else
                return DYN_CHECKSUM_ERROR;
            }
            else
              return error;
          }
          else 
            return DYN_SUCCESS;
        }
        else
          return DYN_SUCCESS;
      }
      else
        return error;
    }
    else
      return DYN_INST_ERROR;
  }
  else 
    return error;
}

unsigned char dyn_master_start_bulk_write(TDynamixelMaster *master,unsigned char num,unsigned char *ids,unsigned short int *address, unsigned short int *length, unsigned char * const data[])
{
  unsigned char error=DYN_COMM_ERROR;

  if(master!=0x00000000)
  {
    // generate the read packet for the desired device
    if(master->version==DYN_VER2)
    {
      dyn2_init_bulk_write_packet(master->tx_buffer,num,ids,address,length,data);
      master->rx_num_packets=0x01;
      master->rx_no_answer=0x00;
      // enable transmission
      master->set_tx_mode();
      // send the data
      if((error=dyn_master_send(master))!=DYN_SUCCESS)
        master->set_rx_mode();
 
      return error;
    }
    else
      return DYN_INST_ERROR;
  }
  else
    return error;
}

unsigned char dyn_master_is_bulk_write_done(TDynamixelMaster *master,unsigned char num,unsigned char *ids,unsigned short int *address, unsigned short int *length, unsigned char * const data[])
{
  unsigned char error=DYN_COMM_ERROR;

  if(master!=0x00000000)
  {
    if(master->version==DYN_VER2)
    {
      error=comm_is_send_done(master->comm_dev);
      if(error==COMM_BUSY)
        return DYN_BUSY;
      else if(error==COMM_SUCCESS)
        return DYN_SUCCESS;
      else
      {
        master->set_rx_mode();
        return DYN_COMM_ERROR;
      }
    }
    else
      return DYN_INST_ERROR;
  }
  else 
    return error;
}

unsigned char dyn_master_start_bulk_read(TDynamixelMaster *master,unsigned char num,unsigned char *ids,unsigned short int *address, unsigned short int *length, unsigned char * const data[])
{
  unsigned char error=DYN_COMM_ERROR,i,ver1_address[255],ver1_length[255];

  if(master!=0x00000000)
  {
    // generate the read packet for the desired device
    if(master->version==DYN_VER1)
    {
      for(i=0;i<num;i++)
      {
        ver1_address[i]=address[i];
        ver1_length[i]=length[i];
      }
      dyn_init_bulk_read_packet(master->tx_buffer,num,ids,ver1_address,ver1_length);
    }
    else
      dyn2_init_bulk_read_packet(master->tx_buffer,num,ids,address,length);
    master->rx_num_packets=num;
    if(master->return_level==no_return)
      master->rx_no_answer=0x01;
    else
      master->rx_no_answer=0x00;
    // enable transmission
    master->set_tx_mode();
    // send the data
    if((error=dyn_master_send(master))!=DYN_SUCCESS)
      master->set_rx_mode();
  }

  return error;
}

unsigned char dyn_master_is_bulk_read_done(TDynamixelMaster *master,unsigned char num,unsigned char *ids,unsigned short int *address, unsigned short int *length, unsigned char * const data[])
{
  static unsigned char num_dev_done=0;
  unsigned char error=DYN_COMM_ERROR;

  if(master!=0x00000000)
  {
    if((error=dyn_master_is_transmission_done(master))==DYN_SUCCESS)
    {
      // wait for the replay within the given timeout
      if(master->return_level!=no_return)
      {
        if(num_dev_done<num)
        {
          if((error=dyn_master_is_reception_done(master))==DYN_SUCCESS)
          {
            // check the input packet checksum
            if(master->version==DYN_VER1)
            {
              if(dyn_check_checksum(master->rx_buffer)==0xFF)
              {
                if((error=dyn_get_status_error(master->rx_buffer))==DYN_SUCCESS)
                {
                  while(dyn_get_read_status_id(master->rx_buffer)!=ids[num_dev_done]) num_dev_done++;
                  if(dyn_get_read_status_data(master->rx_buffer,(unsigned char *)data[num_dev_done])!=length[num_dev_done])// not enough data
                    return DYN_INST_ERROR;
                  else
                  {
                    num_dev_done++;
                    return DYN_BUSY;
                  }
                }
                else
                  return error;
              }
              else
                return DYN_CHECKSUM_ERROR;
            }
            else
            {
              if(dyn2_check_checksum(master->rx_buffer)==0x01)
              {
                if((error=dyn2_get_status_error(master->rx_buffer))==DYN_SUCCESS)
                {
                  while(dyn2_get_read_status_id(master->rx_buffer)!=ids[num_dev_done]) num_dev_done++;
                  if(dyn2_get_read_status_data(master->rx_buffer,(unsigned char *)data[num_dev_done])!=length[num_dev_done])// not enough data
                    return DYN_INST_ERROR;
                  else
                  {
                    num_dev_done++;
                    return DYN_BUSY;
                  }
                }
                else
                  return error;
              }
              else
                return DYN_CHECKSUM_ERROR;
            }
          }
          else
            return error;
        }
        else
          return DYN_SUCCESS;
      }
      else
        return DYN_SUCCESS;
    }
    else
      return error; 
  }
  else 
    return error;
}

/* public functions */
void dyn_master_init(TDynamixelMaster *master,TComm *dev,TDynVersion version)
{
  /* assign communication functions */
  dev->irq_receive_cb=(unsigned char (*)(void *,unsigned char))dyn_master_irq_receive_cb;
  dev->dma_send_cb=(unsigned char (*)(void *))dyn_master_dma_send_cb;
  dev->dma_receive_cb=(unsigned char (*)(void *))dyn_master_dma_receive_cb;
  master->comm_dev=dev;
  master->version=version;
  dev->data=master;
  /* initialize the internal callbacks */
  master->set_tx_mode=dummy_dyn_master_set_tx_mode;
  master->set_rx_mode=dummy_dyn_master_set_rx_mode;
  /* initialize internal variables */
  master->packet_ready=0x00;
  master->received_bytes=0x00;
  master->rx_timeout_ms=50;
  master->rx_no_answer=0x00;
  master->rx_num_packets=0x00;
  master->return_level=return_all;

  master->set_rx_mode();
}

void dyn_master_scan(TDynamixelMaster *master,unsigned char *num,unsigned char *ids)
{
  unsigned char i;

  *num=0;
  if(master!=0x00000000)
  {
    for(i=0;i<254;i++)
    {
      if(dyn_master_ping(master,i)==DYN_SUCCESS)// the device exists
      {
        ids[*num]=i;
        (*num)++;
      }
    }
  }
}

unsigned char dyn_master_ping(TDynamixelMaster *master,unsigned char id)
{
  unsigned char error=DYN_COMM_ERROR;

  if(master!=0x00000000)
  {
    // generate the ping packet for the desired device
    if(master->version==DYN_VER1)
      dyn_init_ping_packet(master->tx_buffer,id);
    else
      dyn2_init_ping_packet(master->tx_buffer,id);
    master->rx_num_packets=0x01;
    master->rx_no_answer=0x00;
    // enable transmission
    master->set_tx_mode();
    // send the data
    if((error=dyn_master_send(master))!=DYN_SUCCESS)
    {
      master->set_rx_mode();
      return error;
    }
    // wait for the transmission to end
    if((error=dyn_master_wait_transmission(master))!=DYN_SUCCESS)
    {
      master->set_rx_mode();
      return error;
    }
    // wait for the replay within the given timeout
    error=dyn_master_wait_reception(master);
  }

  return error;
}

unsigned char dyn_master_read_byte(TDynamixelMaster *master,unsigned char id,unsigned short int address,unsigned char *data)
{
  return dyn_master_read_table(master,id,address,1,data);
}

unsigned char dyn_master_read_word(TDynamixelMaster *master,unsigned char id,unsigned short int address,unsigned short int *data)
{
  unsigned char error;
  unsigned char data_int[2];

  // generate the ping packet for the desired device
  error=dyn_master_read_table(master,id,address,2,data_int);
  (*data)=data_int[0]+data_int[1]*256;

  return error;
}

unsigned char dyn_master_read_table(TDynamixelMaster *master,unsigned char id,unsigned short int address,unsigned short int length,unsigned char *data)
{
  unsigned char error=DYN_COMM_ERROR;

  if(master!=0x00000000)
  {
    // generate the read packet for the desired device
    if(master->version==DYN_VER1)
      dyn_init_read_packet(master->tx_buffer,id,address,length);
    else
      dyn2_init_read_packet(master->tx_buffer,id,address,length);
    master->rx_num_packets=0x01;
    if(master->return_level==no_return || id==DYN_BROADCAST_ID)
      master->rx_no_answer=0x01;
    else
      master->rx_no_answer=0x00;
    // enable transmission
    master->set_tx_mode();
    // send the data
    if((error=dyn_master_send(master))!=DYN_SUCCESS)
    { 
      master->set_rx_mode();
      return error;
    }
    // wait for the transmission to end
    if((error=dyn_master_wait_transmission(master))!=DYN_SUCCESS)
    {
      master->set_rx_mode();
      return error;
    }
    // wait for the replay within the given timeout
    if(master->return_level!=no_return && id!=DYN_BROADCAST_ID)
    {
      if((error=dyn_master_wait_reception(master))==DYN_SUCCESS)
      {
        if(master->version==DYN_VER1)
        {
          if(dyn_get_read_status_data(master->rx_buffer,data)!=length)// not enough data
            error=DYN_INST_ERROR;
        }
        else
        {
          if(dyn2_get_read_status_data(master->rx_buffer,data)!=length)// not enough data
            error=DYN_INST_ERROR;
        }
      }
    } 
  }

  return error;
}

unsigned char dyn_master_write_byte(TDynamixelMaster *master,unsigned char id, unsigned short int address, unsigned char data)
{
  return dyn_master_write_table(master,id,address,1,&data);
}

unsigned char dyn_master_write_word(TDynamixelMaster *master,unsigned char id, unsigned short int address, unsigned short int data)
{
  unsigned char data_int[2];

  data_int[0]=data%256;
  data_int[1]=data/256;
  return dyn_master_write_table(master,id,address,2,data_int);
}

unsigned char dyn_master_write_table(TDynamixelMaster *master,unsigned char id, unsigned short int address, unsigned short int length, unsigned char *data)
{
  unsigned char error=DYN_COMM_ERROR;

  if(master!=0x00000000)
  {
    // generate the write packet for the desired device
    if(master->version==DYN_VER1)
      dyn_init_write_packet(master->tx_buffer,id,address,length,data);
    else
      dyn2_init_write_packet(master->tx_buffer,id,address,length,data);
    master->rx_num_packets=0x01;
    if(master->return_level==return_all && id!=DYN_BROADCAST_ID)
      master->rx_no_answer=0x00;
    else
      master->rx_no_answer=0x01;
    // enable transmission
    master->set_tx_mode();
    // send the data
    if((error=dyn_master_send(master))!=DYN_SUCCESS)
    {
      master->set_rx_mode();
      return error;
    }
    // wait for the transmission to end
    if((error=dyn_master_wait_transmission(master))!=DYN_SUCCESS)
    {
      master->set_rx_mode();
      return error;
    }
    // wait for the replay within the given timeout
    if(master->return_level==return_all && id!=DYN_BROADCAST_ID)
      error=dyn_master_wait_reception(master);
  }

  return error;
}

unsigned char dyn_master_reg_write(TDynamixelMaster *master,unsigned char id, unsigned short int address, unsigned short int length, unsigned char *data)
{
  unsigned char error=DYN_COMM_ERROR;

  if(master!=0x00000000)
  {
    // generate the registered write packet for the desired device
    if(master->version==DYN_VER1)
      dyn_init_reg_write_packet(master->tx_buffer,id,address,length,data);
    else
      dyn2_init_reg_write_packet(master->tx_buffer,id,address,length,data);
    master->rx_num_packets=0x01;
    if(master->return_level==return_all && id!=DYN_BROADCAST_ID)
      master->rx_no_answer=0x00;
    else
      master->rx_no_answer=0x01;
    // enable transmission
    master->set_tx_mode();
    // send the data
    if((error=dyn_master_send(master))!=DYN_SUCCESS)
    {
      master->set_rx_mode();
      return error;
    }
    // wait for the transmission to end
    if((error=dyn_master_wait_transmission(master))!=DYN_SUCCESS)
    {
      master->set_rx_mode();
      return error;
    }
    // wait for the replay within the given timeout
    if(master->return_level==return_all && id!=DYN_BROADCAST_ID)
      error=dyn_master_wait_reception(master);
  }

  return error;
}

unsigned char dyn_master_action(TDynamixelMaster *master)
{
  unsigned char error=DYN_COMM_ERROR;

  if(master!=0x00000000)
  {
    // generate the action packet for the desired device
    if(master->version==DYN_VER1)
      dyn_init_action_packet(master->tx_buffer);
    else
      dyn2_init_action_packet(master->tx_buffer);
    master->rx_num_packets=0x01;
    master->rx_no_answer=0x01;
    // enable transmission
    master->set_tx_mode();
    // send the data
    if((error=dyn_master_send(master))!=DYN_SUCCESS)
    {
      master->set_rx_mode();
      return error;
    }
    // wait for the transmission to end
    if((error=dyn_master_wait_transmission(master))!=DYN_SUCCESS)
    {
      master->set_rx_mode();
      return error;
    }
  }

  return error;
}

unsigned char dyn_master_sync_write(TDynamixelMaster *master,unsigned char num,unsigned char *ids,unsigned short int address, unsigned short int length, unsigned char * const data[])
{
  unsigned char error=DYN_COMM_ERROR;

  if(master!=0x00000000)
  {
    // generate the write packet for the desired device
    if(master->version==DYN_VER1)
      dyn_init_sync_write_packet(master->tx_buffer,num,ids,address,length,data);
    else
      dyn2_init_sync_write_packet(master->tx_buffer,num,ids,address,length,data);
    master->rx_num_packets=0x01;
    master->rx_no_answer=0x01;
    // enable transmission
    master->set_tx_mode();
    // send the data
    if((error=dyn_master_send(master))!=DYN_SUCCESS)
    {
      master->set_rx_mode();
      return error;
    }
    // wait for the transmission to end
    if((error=dyn_master_wait_transmission(master))!=DYN_SUCCESS)
    {
      master->set_rx_mode();
      return error;
    }
  }

  return error;
}

unsigned char dyn_master_sync_read(TDynamixelMaster *master,unsigned char num,unsigned char *ids,unsigned short int address, unsigned short int length, unsigned char * const data[])
{
  unsigned char error=DYN_SUCCESS,i;

  if(master!=0x00000000)
  {
    // generate the read packet for the desired device
    if(master->version==DYN_VER2)
    {
      dyn2_init_sync_read_packet(master->tx_buffer,num,ids,address,length);
      master->rx_num_packets=num;
      if(master->return_level==no_return)
        master->rx_no_answer=0x01;
      else
        master->rx_no_answer=0x00;
      // enable transmission
      master->set_tx_mode();
      // send the data
      if((error=dyn_master_send(master))!=DYN_SUCCESS)
      {
        master->set_rx_mode();
        return error;
      }
      // wait for the transmission to end
      if((error=dyn_master_wait_transmission(master))!=DYN_SUCCESS)
      {
        master->set_rx_mode();
        return error;
      }
      // wait for the replay within the given timeout
      if(master->return_level!=no_return)
      {
        for(i=0;i<num;i++)
        {
          if((error=dyn_master_wait_reception(master))==DYN_SUCCESS)
          {
            while(dyn2_get_read_status_id(master->rx_buffer)!=ids[i]) i++;
            if(dyn2_get_read_status_data(master->rx_buffer,(unsigned char *)data[i])!=length)// not enough data
              error=DYN_INST_ERROR;
          }
        }
      }
    }
  }

  return error;
}

unsigned char dyn_master_bulk_read(TDynamixelMaster *master,unsigned char num,unsigned char *ids,unsigned short int *address, unsigned short int *length, unsigned char * const data[])
{
  unsigned char error=DYN_COMM_ERROR,i,ver1_address[255],ver1_length[255];

  if(master!=0x00000000)
  {
    // generate the read packet for the desired device
    if(master->version==DYN_VER1)
    {
      for(i=0;i<num;i++)
      {
        ver1_address[i]=address[i];
        ver1_length[i]=length[i];
      }
      dyn_init_bulk_read_packet(master->tx_buffer,num,ids,ver1_address,ver1_length);
    }
    else
      dyn2_init_bulk_read_packet(master->tx_buffer,num,ids,address,length);
    master->rx_num_packets=num;
    if(master->return_level==no_return)
      master->rx_no_answer=0x01;
    else
      master->rx_no_answer=0x00;
    // enable transmission
    master->set_tx_mode();
    // send the data
    if((error=dyn_master_send(master))!=DYN_SUCCESS)
    {
      master->set_rx_mode();
      return error;
    }
    // wait for the transmission to end
    if((error=dyn_master_wait_transmission(master))!=DYN_SUCCESS)
    {
      master->set_rx_mode();
      return error;
    }
    // wait for the replay within the given timeout
    if(master->return_level!=no_return)
    {
      for(i=0;i<num;i++)
      {
        if((error=dyn_master_wait_reception(master))==DYN_SUCCESS)
        {
          if(master->version==DYN_VER1)
          {
            while(dyn_get_read_status_id(master->rx_buffer)!=ids[i]) i++;
            if(dyn_get_read_status_data(master->rx_buffer,(unsigned char *)data[i])!=length[i])// not enough data
              error=DYN_INST_ERROR;
          }
          else
          {
            while(dyn2_get_read_status_id(master->rx_buffer)!=ids[i]) i++;
            if(dyn2_get_read_status_data(master->rx_buffer,(unsigned char *)data[i])!=length[i])// not enough data
              error=DYN_INST_ERROR;
          }
        }
      }
    }
  }

  return error;
}

unsigned char dyn_master_bulk_write(TDynamixelMaster *master,unsigned char num,unsigned char *ids,unsigned short int *address, unsigned short int *length, unsigned char * const data[])
{
  unsigned char error=DYN_SUCCESS;

  if(master!=0x00000000)
  {
    // generate the write packet for the desired device
    if(master->version==DYN_VER2)
    {
      dyn2_init_bulk_write_packet(master->tx_buffer,num,ids,address,length,data);
      master->rx_num_packets=0x01;
      master->rx_no_answer=0x01;
      // enable transmission
      master->set_tx_mode();
      // send the data
      if((error=dyn_master_send(master))!=DYN_SUCCESS)
      {
        master->set_rx_mode();
        return error;
      }
      // wait for the transmission to end
      if((error=dyn_master_wait_transmission(master))!=DYN_SUCCESS)
      {
        master->set_rx_mode();
        return error;
      }
    }
  }

  return error;
}

unsigned char dyn_master_relay(TDynamixelMaster *master,unsigned char *inst_pkt,unsigned char *status_pkt)
{
  unsigned char error=DYN_COMM_ERROR;

  if(master!=0x00000000)
  {
    dyn2_copy_packet(inst_pkt,master->tx_buffer);
    master->rx_num_packets=0x01;
    master->rx_no_answer=0x00;
    // enable transmission
    master->set_tx_mode();
    // send the data
    if((error=dyn_master_send(master))!=DYN_SUCCESS)
    {
      master->set_rx_mode();
      return error;
    }
    // wait for the transmission to end
    if((error=dyn_master_wait_transmission(master))!=DYN_SUCCESS)
    {
      master->set_rx_mode();
      return error;
    }
    // wait for the replay within the given timeout
    error=dyn_master_wait_reception(master);
    dyn2_copy_packet(master->rx_buffer,status_pkt);
  }

  return error;
}
