#ifndef _DYN_SERVOS_H
#define _DYN_SERVOS_H

#define NUM_REG 62

typedef struct{
  unsigned short int address;
  unsigned char size;
  unsigned char eeprom;
}TDynReg;

extern TDynReg ax_reg[NUM_REG];
extern TDynReg mx_1_0_reg[NUM_REG];
extern TDynReg mx_2_0_reg[NUM_REG];
extern TDynReg xl_reg[NUM_REG];

typedef enum {
  //                           [Access] [Description]
  // Info
  model_number=0,      //   (R)     Model Number
  model_info,          //   (R)
  firmware_version,    //   (R)     Version of the Firmware
  id,                  //   (RW)    ID of Dynamixel
  baudrate,            //   (RW)    Baud Rate of Dynamixel
  return_delay_time,   //   (RW)    Return Delay Time 
  second_id,           //   (RW)
  protocol,            //   (RW)
  return_level,        //   (RW)    Status Return Level
  // Configuration
  drive_mode,          //   (RW)    Drive mode
  op_mode,             //   (RW)    Joint or wheel mode
  homming_offset,      //   (RW)
  moving_threshold,    //   (RW)
  min_angle_limit,     //   (RW)    Clockwise Angle Limit
  max_angle_limit,     //   (RW)    Counterclockwise Angle Limit 
  temp_limit,          //   (RW)    Internal Temperature Limit
  min_voltage_limit,   //   (RW)    Lowest Voltage Limit
  max_voltage_limit,   //   (RW)    Highest Voltage Limit
  pwm_limit,           //   (RW)    Maximum Torque 
  max_pwm,             //   (RW)    Maximum Torque 
  current_limit,       //   (RW)
  max_current,         //   (RW)
  velocity_limit,      //   (RW)
  multi_turn_off,      //   (RW)    Multi turn offset position
  resolution_div,      //   (RW)    Resolution divider
  // Status
  alarm_led,           //   (RW)    LED for Alarm 
  alarm_shtdwn,        //   (RW)    Shutdown for Alarm 
  led,                 //   (RW)    LED On/Off 
  bus_watchdog,        //   (RW)
  moving,              //   (R)     Means if there is any movement 
  moving_status,       //   (R)
  lock,                //   (RW)    Locking EEPROM 
  reg_inst,            //   (RW)    Means if Instruction is registered 
  hardware_error,      //   (R)     Means if there is any movement 
  // control
  torque_en,           //   (RW)    Torque On/Off 
  vel_pid_i,           //   (RW)
  vel_pid_p,           //   (RW)
  pos_pid_p,           //   (RW)     
  pos_pid_i,           //   (RW)     
  pos_pid_d,           //   (RW)     
  feedfwd_gain_2,      //   (RW)
  feedfwd_gain_1,      //   (RW)
  cw_comp_margin,      //   (RW)    CW Compliance Margin
  ccw_comp_margin,     //   (RW)    CCW Compliance Margin
  cw_comp_slope,       //   (RW)    CW Compliance Slope
  ccw_comp_slope,      //   (RW)    CCW Compliance Slope 
  punch,               //   (RW)    Punch 
  // Set point
  goal_pos,            //   (RW)    Goal Position 
  goal_speed,          //   (RW)    Moving Speed 
  goal_pwm,            //   (RW)
  goal_current,        //   (RW)
  profile_accel,       //   (RW)
  profile_vel,         //   (RW)
  // Feedback
  current_pos,         //   (R)     Current Position 
  current_speed,       //   (R)     Current Speed 
  current_load,        //   (R)     Current Load 
  current_voltage,     //   (R)     Current Voltage 
  current_temp,        //   (R)     Current Temperature 
  current_pwm,         //   (R)
  realtime_tick,       //   (R)     Realtime tick
  velocity_traj,       //   (R)
  pos_traj             //   (R)
} reg_id;

#define NUM_SERVO_MODELS 11

typedef struct{
  unsigned short int model;
  unsigned short int encoder_resolution;
  unsigned char gear_ratio;
  unsigned short int max_angle;
  unsigned short int center_angle;
  unsigned short int max_speed;
  long long int cw_angle_limit;
  long long int ccw_angle_limit;
  TDynReg *registers;
  unsigned char pid;
  unsigned char protocol_ver;
}TServoConfig;

extern TServoConfig servo_data[NUM_SERVO_MODELS];

#endif
