#include "dyn_module.h"
#include "dyn_module_registers.h"
#include "ram.h"

// private functions
void dyn_module_write_cmd(TDynModule *module,unsigned short int address,unsigned short int length,unsigned char *data)
{
  unsigned short int eeprom_offset;

  eeprom_offset=address-module->eeprom_base_address;
  if(ram_in_window(module->eeprom_base_address+DYN_MODULE_PERIOD_OFFSET,1,address,length))
    dyn_module_set_period(module,data[DYN_MODULE_PERIOD_OFFSET-eeprom_offset]);
  ram_write_table(module->memory,address,length,data);
}

void dyn_module_read_cmd(TDynModule *module,unsigned short int address,unsigned short int length,unsigned char *data)
{
  ram_read_table(module->memory,address,length,data);
}

// public functions
unsigned char dyn_module_init(TDynModule *module,TMemory *memory,unsigned short int ram_base_address,unsigned short int eeprom_base_address)
{
  unsigned char i;

  module->num_models=0x00;
  for(i=0;i<MODULE_MAX_NUM_MODELS;i++)
    module->models[i]=0x0000;
  for(i=0;i<DYN_MANAGER_MAX_NUM_DEVICES;i++)
    module->assigned_ids[i]=0x00;
  module->num_assigned_ids=0x00;
  module->period_count=0x00;
  module->enabled=0x01;
  module->add_device=0x00000000;
  module->set_period=0x00000000;
  module->setup=0x00000000;
  module->pre_process=0x00000000;
  module->post_process=0x00000000;
  module->data=0x00000000;
  /* initialize memory module */
  mem_module_init(&module->mem_module);
  module->mem_module.data=module;
  module->mem_module.write_cmd=(void(*)(void *,unsigned short int, unsigned short int,unsigned char *))dyn_module_write_cmd;
  module->mem_module.read_cmd=(void(*)(void *,unsigned short int, unsigned short int,unsigned char *))dyn_module_read_cmd;
  if(!mem_module_add_ram_segment(&module->mem_module,ram_base_address,RAM_DYN_MODULE_LENGTH))
    return 0x00;
  if(!mem_module_add_eeprom_segment(&module->mem_module,eeprom_base_address,EEPROM_DYN_MODULE_LENGTH))
    return 0x00;
  if(!mem_add_module(memory,&module->mem_module))
    return 0x00;
  module->eeprom_base_address=eeprom_base_address;
  module->ram_base_address=ram_base_address;
  module->memory=memory;

  return 0x01;
}

void dyn_module_add_model(TDynModule *module,unsigned short int model)
{
  unsigned char i;

  for(i=0;i<module->num_models;i++)
    if(module->models[i]==model)
      return;
  if(i<MODULE_MAX_NUM_MODELS)
  {
    module->models[i]=model;
    module->num_models++;
  }
}

void dyn_module_add_device(TDynModule *module,unsigned char id, unsigned short int model)
{
  unsigned char i;
  
  for(i=0;i<module->num_models;i++)
  {
    if(model==module->models[i])
    {
      module->assigned_ids[id]=0x01;
      module->num_assigned_ids++;
      if(module->add_device!=0x00000000)
        module->add_device(module->data,id,model);
      break;
    }
  }
}

unsigned char dyn_module_device_is_assigned(TDynModule *module,unsigned char id)
{
  if(id<DYN_MANAGER_MAX_NUM_DEVICES)
    return module->assigned_ids[id];
  else
    return 0x00;
}

void dyn_module_set_period(TDynModule *module,unsigned char period)
{
  if(period==0x00)
    period=0x01;
  module->period_count=period;
  if(module->set_period!=0x00000000)
    module->set_period(module->data,period*module->manager->period_ms);
}
