#include "joint_motion.h"
#include "joint_motion_mm_registers.h"
#include "ram.h"
#include <stdlib.h>

// private functions
void joint_motion_read_cmd(TJointMModule *module,unsigned short int address,unsigned short int length,unsigned char *data)
{
  ram_read_table(module->memory,address,length,data);
}

void joint_motion_write_cmd(TJointMModule *module,unsigned short int address,unsigned short int length,unsigned char *data)
{
  unsigned char j;
  unsigned short int i;
  unsigned int servos,ram_offset;

  // load motion information
  ram_offset=address-module->ram_base_address;
  for(i=0;i<DYN_MANAGER_MAX_NUM_DEVICES*4;i++)
    if(ram_in_range(module->ram_base_address+JOINT_MOTION_MM_SERVO_ANGLES_OFFSET+i,address,length))
      module->memory->data[i]=data[JOINT_MOTION_MM_SERVO_ANGLES_OFFSET+i-ram_offset];
  // group servos
  for(j=0;j<NUM_JOINT_GROUPS;j++)
  {
    if(ram_in_window(module->ram_base_address+JOINT_MOTION_MM_GROUP_SERVOS_OFFSET+j*4,4,address,length))
    {
      servos=joint_motion_get_group_servos(module,j);
      for(i=0;i<4;i++)
        if(ram_in_range(module->ram_base_address+JOINT_MOTION_MM_GROUP_SERVOS_OFFSET+j*4+i,address,length))
          servos|=(data[JOINT_MOTION_MM_GROUP_SERVOS_OFFSET+j*4+i-ram_offset]<<(i*8));
      joint_motion_set_group_servos(module,j,servos);
    }
    if(ram_in_range(module->ram_base_address+JOINT_MOTION_MM_GROUP_CONTROL_OFFSET+j,address,length))
    {
      if(data[JOINT_MOTION_MM_GROUP_CONTROL_OFFSET+j-ram_offset]&JOINT_MOTION_MM_START)
        joint_motion_start(module,j);
      if(data[JOINT_MOTION_MM_GROUP_CONTROL_OFFSET+j-ram_offset]&JOINT_MOTION_MM_STOP)
        joint_motion_stop(module,j);
    }
  }
}

TMotionModule *joint_motion_get_module(TJointMModule *joint)
{
  return &joint->mmodule;
}

void joint_motion_set_period(void *module,unsigned short int period_ms)
{
  TJointMModule *joint=(TJointMModule *)module;

  // load the current period 
  joint->motion_period=(period_ms<<16)/1000;
  joint->motion_period_ms=period_ms;
}

void joint_motion_set_module(void *module,unsigned char servo_id)
{

}

// motion manager interface functions
void joint_motion_pre_process(void *module)
{
  TJointMModule *joint=(TJointMModule *)module;
  unsigned char moving,i,j;
  unsigned int servo_index;

  for(j=0;j<NUM_JOINT_GROUPS;j++)
  {
    moving=0x00;
    if(joint->moving[j]==0x01)
    {
      for(i=0,servo_index=0x00000001;i<DYN_MANAGER_MAX_NUM_DEVICES;i++,servo_index=servo_index<<1)
      {
        if(mmanager_get_module(joint->mmodule.manager,i)==MM_JOINTS && joint->group_servos[j]&servo_index)
        {
          if(joint->stop[j]==0x01)
          {
            joint->target_angles[i]=joint->mmodule.manager->servo_values[i].target_angle;
            joint->target_speeds[i]=0;
            joint->target_accels[i]=0;
            joint->current_speeds[i]=0;
          }
          else
          {
            if(abs(joint->target_angles[i]-joint->current_angles[i])>=65)
            {
              moving=0x01;
              if(joint->current_speeds[i]!=joint->dir[i]*joint->target_speeds[i])// it is necessary to change the current speed
              {
                joint->current_angles[i]+=((joint->current_speeds[i]*joint->motion_period)>>17);
                if(joint->current_speeds[i]>joint->dir[i]*joint->target_speeds[i])
                {
                  joint->current_speeds[i]-=((joint->target_accels[i]*joint->motion_period)>>16);// reduce speed
                  if(joint->current_speeds[i]<joint->dir[i]*joint->target_speeds[i])
                    joint->current_speeds[i]=joint->dir[i]*joint->target_speeds[i];
                }
                else
                {
                  joint->current_speeds[i]+=((joint->target_accels[i]*joint->motion_period)>>16);// increase speed
                  if(joint->current_speeds[i]>joint->dir[i]*joint->target_speeds[i])
                    joint->current_speeds[i]=joint->dir[i]*joint->target_speeds[i];
                }
                joint->current_angles[i]+=((joint->current_speeds[i]*joint->motion_period)>>17);
                if(joint->dir[i]==1 && joint->current_angles[i]>joint->target_angles[i])
                  joint->current_angles[i]=joint->target_angles[i];
                if(joint->dir[i]==-1 && joint->current_angles[i]<joint->target_angles[i])
                  joint->current_angles[i]=joint->target_angles[i];
              }  
              else
              {
                if(abs(joint->target_angles[i]-joint->current_angles[i])>joint->target_stop_angles[i])// constant speed phase
                {
                  joint->current_angles[i]+=((joint->current_speeds[i]*joint->motion_period)>>16);
                  if(joint->dir[i]==1 && joint->current_angles[i]>joint->target_angles[i])
                    joint->current_angles[i]=joint->target_angles[i];
                  if(joint->dir[i]==-1 && joint->current_angles[i]<joint->target_angles[i])
                    joint->current_angles[i]=joint->target_angles[i];
                }
                else// deceleration phase
                {
                  joint->current_angles[i]=joint->current_angles[i]+((joint->current_speeds[i]*joint->motion_period)>>17);
                  joint->target_speeds[i]=joint->target_speeds[i]-((joint->target_accels[i]*joint->motion_period)>>16);
                  if(joint->target_speeds[i]<0)
                    joint->target_speeds[i]=0;
                  joint->current_speeds[i]=joint->dir[i]*joint->target_speeds[i];
                  joint->current_angles[i]=joint->current_angles[i]+((joint->current_speeds[i]*joint->motion_period)>>17);
                  if(joint->dir[i]==1 && joint->current_angles[i]>joint->target_angles[i])
                    joint->current_angles[i]=joint->target_angles[i];
                  if(joint->dir[i]==-1 && joint->current_angles[i]<joint->target_angles[i])
                    joint->current_angles[i]=joint->target_angles[i];
                }
              }
            }
            else
            {
              joint->current_speeds[i]=0;
            }
            joint->mmodule.manager->servo_values[i].target_angle=joint->current_angles[i];
            if(joint->mmodule.manager->servo_configs[i]->pid)
            {
            }
            else
            {
              joint->mmodule.manager->servo_values[i].cw_compliance=5;
              joint->mmodule.manager->servo_values[i].ccw_compliance=5;
            }
          }
        }
      }
      if(moving==0)
      {
        joint->moving[j]=0x00;
        joint->stop[j]=0x00;
        joint->memory->data[joint->ram_base_address+JOINT_MOTION_MM_GROUP_CONTROL_OFFSET+j]&=(~JOINT_MOTION_MM_RUNNING);
      }
    }
  }
}

// public functions
unsigned char joint_motion_init(TJointMModule *joint,TMemory *memory,unsigned short int ram_base_address)
{
  unsigned char i;

  /* initialize the internal variables */
  for(i=0;i<DYN_MANAGER_MAX_NUM_DEVICES;i++)
  {
    // target values
    joint->target_angles[i]=0;
    joint->target_speeds[i]=0;
    joint->target_accels[i]=0;
    // current values
    joint->current_angles[i]=0;
    joint->current_speeds[i]=0;
    // the current angles, speeds and accelerations are in RAM
    joint->dir[i]=0;
  }
  for(i=0;i<NUM_JOINT_GROUPS;i++)
  {
    joint->stop[i]=0x00;
    joint->moving[i]=0x00;
    joint->group_servos[i]=0x00000000;
  }

  /* initialize the motion module */
  mmodule_init(&joint->mmodule);
  joint->mmodule.id=MM_JOINTS;
  joint->mmodule.set_period=joint_motion_set_period;
  joint->mmodule.set_module=joint_motion_set_module;
  joint->mmodule.pre_process=joint_motion_pre_process;
  joint->mmodule.data=joint;

  mem_module_init(&joint->mem_module);
  joint->mem_module.data=joint;
  joint->mem_module.write_cmd=(void(*)(void *,unsigned short int, unsigned short int,unsigned char *))joint_motion_write_cmd;
  joint->mem_module.read_cmd=(void(*)(void *,unsigned short int, unsigned short int,unsigned char *))joint_motion_read_cmd;
  if(!mem_module_add_ram_segment(&joint->mem_module,ram_base_address,RAM_JOINT_MOTION_MM_LENGTH))
    return 0x00;
  if(!mem_add_module(memory,&joint->mem_module))
    return 0x00;
  joint->ram_base_address=ram_base_address;
  joint->memory=memory;

  return 0x01;
}

void joint_motion_set_group_servos(TJointMModule *joint,unsigned char group_id,unsigned int servos)
{
  joint->group_servos[group_id]=servos;
  joint->memory->data[joint->ram_base_address+JOINT_MOTION_MM_GROUP_SERVOS_OFFSET+group_id*4]=servos&0x000000FF;
  joint->memory->data[joint->ram_base_address+JOINT_MOTION_MM_GROUP_SERVOS_OFFSET+group_id*4+1]=(servos>>8)&0x000000FF;
  joint->memory->data[joint->ram_base_address+JOINT_MOTION_MM_GROUP_SERVOS_OFFSET+group_id*4+2]=(servos>>16)&0x000000FF;
  joint->memory->data[joint->ram_base_address+JOINT_MOTION_MM_GROUP_SERVOS_OFFSET+group_id*4+3]=(servos>>24)&0x000000FF;
}

unsigned int joint_motion_get_group_servos(TJointMModule *joint,unsigned char group_id)
{
  return joint->group_servos[group_id];
}

void joint_motion_start(TJointMModule *joint,unsigned char group_id)
{
  unsigned char i;
  short int angle;
  unsigned int servo_index;

  /* initialize the internal variables */
  for(i=0,servo_index=0x00000001;i<DYN_MANAGER_MAX_NUM_DEVICES;i++,servo_index=servo_index<<1)
  {
    if(joint->group_servos[group_id]&servo_index)
    {
      if(joint->memory->data[joint->ram_base_address+JOINT_MOTION_MM_SERVO_SPEEDS_OFFSET+i]==0)
      {
        joint->target_angles[i]=joint->mmodule.manager->servo_values[i].target_angle;
        joint->current_angles[i]=joint->mmodule.manager->servo_values[i].target_angle;
        angle=joint->mmodule.manager->servo_values[i].target_angle>>9;
        joint->memory->data[joint->ram_base_address+JOINT_MOTION_MM_SERVO_ANGLES_OFFSET+i*2]=angle&0x00FF;
        joint->memory->data[joint->ram_base_address+JOINT_MOTION_MM_SERVO_ANGLES_OFFSET+i*2+1]=(angle>>8)&0x00FF;
      }
      else
      {
        // current values
        joint->current_angles[i]=joint->mmodule.manager->servo_values[i].target_angle;
        // target angle
        angle=joint->memory->data[joint->ram_base_address+JOINT_MOTION_MM_SERVO_ANGLES_OFFSET+i*2]+(joint->memory->data[joint->ram_base_address+JOINT_MOTION_MM_SERVO_ANGLES_OFFSET+i*2+1]<<8);
        if(angle>joint->mmodule.manager->servo_configs[i]->ccw_angle_limit)
          angle=joint->mmodule.manager->servo_configs[i]->ccw_angle_limit;
        else if(angle<joint->mmodule.manager->servo_configs[i]->cw_angle_limit)
          angle=joint->mmodule.manager->servo_configs[i]->ccw_angle_limit;
        joint->target_angles[i]=angle<<9;
        // target speed
        joint->target_speeds[i]=joint->memory->data[joint->ram_base_address+JOINT_MOTION_MM_SERVO_SPEEDS_OFFSET+i]<<16;
        // target accel
        joint->target_accels[i]=joint->memory->data[joint->ram_base_address+JOINT_MOTION_MM_SERVO_ACCELS_OFFSET+i]<<16;
        if(joint->target_accels[i]==0)
        {
          joint->target_accels[i]=1<<16;
          joint->memory->data[joint->ram_base_address+JOINT_MOTION_MM_SERVO_ACCELS_OFFSET+i]=1;
        }
        // check the parameters
        if(abs(joint->target_angles[i]-joint->current_angles[i])>=65)
        {
          if((joint->target_speeds[i]*joint->target_speeds[i])/joint->target_accels[i]>abs(joint->target_angles[i]-joint->current_angles[i]))
          {
            joint->target_accels[i]=(joint->target_speeds[i]*joint->target_speeds[i])/abs(joint->target_angles[i]-joint->current_angles[i]);
            joint->memory->data[joint->ram_base_address+JOINT_MOTION_MM_SERVO_ACCELS_OFFSET+i]=joint->target_accels[i]>>16;
          }
        }
        // stop angles
        joint->target_stop_angles[i]=joint->target_speeds[i]*joint->target_speeds[i]/(2*joint->target_accels[i]);
        // the current angles, speeds and accelerations are in RAM
        if(joint->target_angles[i]>=joint->current_angles[i])
          joint->dir[i]=1;
        else
          joint->dir[i]=-1;
      }
    }
  }
  joint->moving[group_id]=0x01;
  joint->stop[group_id]=0x00;
  joint->memory->data[joint->ram_base_address+JOINT_MOTION_MM_GROUP_CONTROL_OFFSET+group_id]|=JOINT_MOTION_MM_RUNNING;
}

void joint_motion_stop(TJointMModule *joint,unsigned char group_id)
{
  joint->stop[group_id]=0x01;
}

unsigned char are_joints_moving(TJointMModule *joint,unsigned char group_id)
{
  return joint->moving[group_id];
}
