#include "darwin_dyn_master_v2.h"
#include "dyn_servos.h"
#include "grippers.h"
#include "ram.h"
#include "joint_motion.h"

// private variables
int64_t grippers_period;
int16_t grippers_period_us;
uint8_t gripper_left_target;
uint8_t gripper_right_target;

// public functions
void grippers_init(uint16_t period_us)
{
  uint16_t force;
  uint32_t servo_mask=0x00000000;

  /* initialize private variables */
  gripper_left_target=0x00;//close;
  gripper_right_target=0x00;//close;
  grippers_period=(period_us<<16)/1000000;
  grippers_period_us=period_us;
  /* configure the maximum force of the servos */
  servo_mask=0x00000000;
  if(ram_data[DARWIN_GRIPPER_LEFT_TOP_ID]!=0xFF && ram_data[DARWIN_GRIPPER_LEFT_BOT_ID]!=0xFF)
  {
    force=ram_data[DARWIN_GRIPPER_LEFT_MAX_FORCE_L]+(ram_data[DARWIN_GRIPPER_LEFT_MAX_FORCE_H]<<8);
    dyn_master_write_word(&darwin_dyn_master_v2,ram_data[DARWIN_GRIPPER_LEFT_TOP_ID],XL_TORQUE_LIMIT_L,force);
    dyn_master_write_word(&darwin_dyn_master_v2,ram_data[DARWIN_GRIPPER_LEFT_BOT_ID],XL_TORQUE_LIMIT_L,force);
    servo_mask|=1<<ram_data[DARWIN_GRIPPER_LEFT_TOP_ID];
    servo_mask|=1<<ram_data[DARWIN_GRIPPER_LEFT_BOT_ID];
    joint_motion_set_group_servos(1,servo_mask);
  }
  servo_mask=0x00000000;
  if(ram_data[DARWIN_GRIPPER_RIGHT_TOP_ID]!=0xFF && ram_data[DARWIN_GRIPPER_RIGHT_BOT_ID]!=0xFF)
  {
    force=ram_data[DARWIN_GRIPPER_RIGHT_MAX_FORCE_L]+(ram_data[DARWIN_GRIPPER_RIGHT_MAX_FORCE_H]<<8);
    dyn_master_write_word(&darwin_dyn_master_v2,ram_data[DARWIN_GRIPPER_RIGHT_TOP_ID],XL_TORQUE_LIMIT_L,force);
    dyn_master_write_word(&darwin_dyn_master_v2,ram_data[DARWIN_GRIPPER_RIGHT_BOT_ID],XL_TORQUE_LIMIT_L,force);
    servo_mask|=1<<ram_data[DARWIN_GRIPPER_RIGHT_TOP_ID];
    servo_mask|=1<<ram_data[DARWIN_GRIPPER_RIGHT_BOT_ID];
    joint_motion_set_group_servos(2,servo_mask);
  }
  grippers_close(GRIPPER_LEFT);
  grippers_close(GRIPPER_RIGHT);
}

void grippers_set_period(uint16_t period_us)
{
  grippers_period=(period_us<<16)/1000000;
  grippers_period_us=period_us;
}

uint16_t grippers_get_period(void)
{
  return grippers_period_us;
}

void grippers_open(grippers_t gripper)
{
  if(gripper==GRIPPER_LEFT)
  {
    ram_data[DARWIN_JOINT_SERVO0_ANGLE_L+ram_data[DARWIN_GRIPPER_LEFT_TOP_ID]*2]=ram_data[DARWIN_GRIPPER_RIGHT_MAX_ANGLE_L];
    ram_data[DARWIN_JOINT_SERVO0_ANGLE_L+ram_data[DARWIN_GRIPPER_LEFT_TOP_ID]*2+1]=ram_data[DARWIN_GRIPPER_RIGHT_MAX_ANGLE_H];
    ram_data[DARWIN_JOINT_SERVO0_SPEED+ram_data[DARWIN_GRIPPER_LEFT_TOP_ID]]=60;
    ram_data[DARWIN_JOINT_SERVO0_ANGLE_L+ram_data[DARWIN_GRIPPER_LEFT_BOT_ID]*2]=ram_data[DARWIN_GRIPPER_RIGHT_MAX_ANGLE_L];
    ram_data[DARWIN_JOINT_SERVO0_ANGLE_L+ram_data[DARWIN_GRIPPER_LEFT_BOT_ID]*2+1]=ram_data[DARWIN_GRIPPER_RIGHT_MAX_ANGLE_H];
    ram_data[DARWIN_JOINT_SERVO0_SPEED+ram_data[DARWIN_GRIPPER_LEFT_BOT_ID]]=60;
    joint_motion_start(1);
  }
  else if(gripper==GRIPPER_RIGHT)
  {
    ram_data[DARWIN_JOINT_SERVO0_ANGLE_L+ram_data[DARWIN_GRIPPER_RIGHT_TOP_ID]*2]=ram_data[DARWIN_GRIPPER_RIGHT_MAX_ANGLE_L];
    ram_data[DARWIN_JOINT_SERVO0_ANGLE_L+ram_data[DARWIN_GRIPPER_RIGHT_TOP_ID]*2+1]=ram_data[DARWIN_GRIPPER_RIGHT_MAX_ANGLE_H];
    ram_data[DARWIN_JOINT_SERVO0_SPEED+ram_data[DARWIN_GRIPPER_RIGHT_TOP_ID]]=60;
    ram_data[DARWIN_JOINT_SERVO0_ANGLE_L+ram_data[DARWIN_GRIPPER_RIGHT_BOT_ID]*2]=ram_data[DARWIN_GRIPPER_RIGHT_MAX_ANGLE_L];
    ram_data[DARWIN_JOINT_SERVO0_ANGLE_L+ram_data[DARWIN_GRIPPER_RIGHT_BOT_ID]*2+1]=ram_data[DARWIN_GRIPPER_RIGHT_MAX_ANGLE_H];
    ram_data[DARWIN_JOINT_SERVO0_SPEED+ram_data[DARWIN_GRIPPER_RIGHT_BOT_ID]]=60;
    joint_motion_start(2);
  }
}

void grippers_close(grippers_t gripper)
{
  if(gripper==GRIPPER_LEFT)
  {
    ram_data[DARWIN_JOINT_SERVO0_ANGLE_L+ram_data[DARWIN_GRIPPER_LEFT_TOP_ID]*2]=ram_data[DARWIN_GRIPPER_RIGHT_MIN_ANGLE_L];
    ram_data[DARWIN_JOINT_SERVO0_ANGLE_L+ram_data[DARWIN_GRIPPER_LEFT_TOP_ID]*2+1]=ram_data[DARWIN_GRIPPER_RIGHT_MIN_ANGLE_H];
    ram_data[DARWIN_JOINT_SERVO0_SPEED+ram_data[DARWIN_GRIPPER_LEFT_TOP_ID]]=60;
    ram_data[DARWIN_JOINT_SERVO0_ANGLE_L+ram_data[DARWIN_GRIPPER_LEFT_BOT_ID]*2]=ram_data[DARWIN_GRIPPER_RIGHT_MIN_ANGLE_L];
    ram_data[DARWIN_JOINT_SERVO0_ANGLE_L+ram_data[DARWIN_GRIPPER_LEFT_BOT_ID]*2+1]=ram_data[DARWIN_GRIPPER_RIGHT_MIN_ANGLE_H];
    ram_data[DARWIN_JOINT_SERVO0_SPEED+ram_data[DARWIN_GRIPPER_LEFT_BOT_ID]]=60;
    joint_motion_start(1);
  }
  else if(gripper==GRIPPER_RIGHT)
  {
    ram_data[DARWIN_JOINT_SERVO0_ANGLE_L+ram_data[DARWIN_GRIPPER_RIGHT_TOP_ID]*2]=ram_data[DARWIN_GRIPPER_RIGHT_MIN_ANGLE_L];
    ram_data[DARWIN_JOINT_SERVO0_ANGLE_L+ram_data[DARWIN_GRIPPER_RIGHT_TOP_ID]*2+1]=ram_data[DARWIN_GRIPPER_RIGHT_MIN_ANGLE_H];
    ram_data[DARWIN_JOINT_SERVO0_SPEED+ram_data[DARWIN_GRIPPER_RIGHT_TOP_ID]]=60;
    ram_data[DARWIN_JOINT_SERVO0_ANGLE_L+ram_data[DARWIN_GRIPPER_RIGHT_BOT_ID]*2]=ram_data[DARWIN_GRIPPER_RIGHT_MIN_ANGLE_L];
    ram_data[DARWIN_JOINT_SERVO0_ANGLE_L+ram_data[DARWIN_GRIPPER_RIGHT_BOT_ID]*2+1]=ram_data[DARWIN_GRIPPER_RIGHT_MIN_ANGLE_H];
    ram_data[DARWIN_JOINT_SERVO0_SPEED+ram_data[DARWIN_GRIPPER_RIGHT_BOT_ID]]=60;
    joint_motion_start(2);
  }
}

void grippers_stop(grippers_t gripper)
{
  if(gripper==GRIPPER_LEFT)
    joint_motion_stop(1);
  else
    joint_motion_stop(2);
}

uint8_t gripper_is_open(grippers_t gripper)
{
  if(gripper==GRIPPER_LEFT)
  {
    if(ram_data[DARWIN_GRIPPER_CNTRL]&GRIPPER_OPENED_LEFT)
      return 0x01;
    else
      return 0x00;
  }
  else
  {
    if(ram_data[DARWIN_GRIPPER_CNTRL]&GRIPPER_OPENED_RIGHT)
      return 0x01;
    else
      return 0x00;
  }
}

uint8_t gripper_is_close(grippers_t gripper)
{
  if(gripper==GRIPPER_LEFT)
  {
    if(ram_data[DARWIN_GRIPPER_CNTRL]&GRIPPER_OPENED_LEFT)
      return 0x00;
    else
      return 0x01;
  }
  else
  {
    if(ram_data[DARWIN_GRIPPER_CNTRL]&GRIPPER_OPENED_RIGHT)
      return 0x00;
    else
      return 0x01;
  }
}

uint8_t gripper_is_moving(grippers_t gripper)
{
  if(gripper==GRIPPER_LEFT)
    return are_joints_moving(1);
  else
    return are_joints_moving(2);
}

// operation functions
uint8_t grippers_in_range(unsigned short int address, unsigned short int length)
{
  if(ram_in_window(GRIPPER_BASE_ADDRESS,GRIPPER_MEM_LENGTH,address,length) ||
     ram_in_window(GRIPPER_EEPROM_ADDRESS,GRIPPER_EEPROM_LENGTH,address,length))
    return 0x01;
  else
    return 0x00;
}

void grippers_process_read_cmd(unsigned short int address,unsigned short int length,unsigned char *data)
{

}

void grippers_process_write_cmd(unsigned short int address,unsigned short int length,unsigned char *data)
{
  if(ram_in_range(DARWIN_GRIPPER_CNTRL,address,length))
  {
    if(data[DARWIN_GRIPPER_CNTRL-address]&GRIPPER_OPEN_RIGHT)
      grippers_open(GRIPPER_RIGHT);
    else if(data[DARWIN_GRIPPER_CNTRL-address]&GRIPPER_CLOSE_RIGHT)
      grippers_close(GRIPPER_RIGHT);
    if(data[DARWIN_GRIPPER_CNTRL-address]&GRIPPER_OPEN_LEFT)
      grippers_open(GRIPPER_LEFT);
    else if(data[DARWIN_GRIPPER_CNTRL-address]&GRIPPER_CLOSE_LEFT)
      grippers_close(GRIPPER_LEFT);
  }
}

void grippers_process(void)
{
  if(!are_joints_moving(1))
  {
    ram_data[DARWIN_GRIPPER_CNTRL]&=~GRIPPER_MOVING_LEFT;
    if(gripper_left_target==0x01)
      ram_data[DARWIN_GRIPPER_CNTRL]|=GRIPPER_OPENED_LEFT;
    else
      ram_data[DARWIN_GRIPPER_CNTRL]&=~GRIPPER_OPENED_LEFT;
  }
  else
  {
    ram_data[DARWIN_GRIPPER_CNTRL]&=~GRIPPER_OPENED_LEFT;
    ram_data[DARWIN_GRIPPER_CNTRL]|=GRIPPER_MOVING_LEFT;
  }
  if(!are_joints_moving(2))
  {
    ram_data[DARWIN_GRIPPER_CNTRL]&=~GRIPPER_MOVING_RIGHT;
    if(gripper_right_target==0x01)
      ram_data[DARWIN_GRIPPER_CNTRL]|=GRIPPER_OPENED_RIGHT;
    else
      ram_data[DARWIN_GRIPPER_CNTRL]&=~GRIPPER_OPENED_RIGHT;
  }
  else
  {
    ram_data[DARWIN_GRIPPER_CNTRL]&=~GRIPPER_OPENED_RIGHT;
    ram_data[DARWIN_GRIPPER_CNTRL]|=GRIPPER_MOVING_RIGHT;
  }
}
