#include "head_tracking.h"
#include "ram.h"

// private variables
typedef struct{
  uint16_t kp;
  uint16_t ki;
  uint16_t kd;
  int64_t prev_error;// 48|16
  int64_t integral_part;
  int64_t integral_clamp;
  int64_t max_angle;
  int64_t min_angle;
  int64_t target_angle;
}TJointControl;

TJointControl head_tracking_pan;
TJointControl head_tracking_tilt;
int64_t head_tracking_period;
int16_t head_tracking_period_us;
uint8_t head_tracking_enabled;

// public functions
void head_tracking_init(uint16_t period_us)
{
  /* initialize private variables */
  head_tracking_pan.kp=ram_data[DARWIN_HEAD_PAN_P_L]+(ram_data[DARWIN_HEAD_PAN_P_H]<<8);
  head_tracking_pan.ki=ram_data[DARWIN_HEAD_PAN_I_L]+(ram_data[DARWIN_HEAD_PAN_I_H]<<8);
  head_tracking_pan.kd=ram_data[DARWIN_HEAD_PAN_D_L]+(ram_data[DARWIN_HEAD_PAN_D_H]<<8);
  head_tracking_pan.prev_error=0;
  head_tracking_pan.integral_part=0;
  head_tracking_pan.integral_clamp=(ram_data[DARWIN_HEAD_PAN_I_CLAMP_L]+(ram_data[DARWIN_HEAD_PAN_I_CLAMP_H]<<8))<<9;
  head_tracking_pan.min_angle=manager_get_cw_angle_limit(L_PAN)<<9;
  ram_data[DARWIN_HEAD_MIN_PAN_L]=manager_get_cw_angle_limit(L_PAN)%256;
  ram_data[DARWIN_HEAD_MIN_PAN_H]=manager_get_cw_angle_limit(L_PAN)/256;
  head_tracking_pan.max_angle=manager_get_ccw_angle_limit(L_PAN)<<9;
  ram_data[DARWIN_HEAD_MAX_PAN_L]=manager_get_ccw_angle_limit(L_PAN)%256;
  ram_data[DARWIN_HEAD_MAX_PAN_H]=manager_get_ccw_angle_limit(L_PAN)/256;
  head_tracking_pan.target_angle=0;
  head_tracking_tilt.kp=ram_data[DARWIN_HEAD_TILT_P_L]+(ram_data[DARWIN_HEAD_TILT_P_H]<<8);
  head_tracking_tilt.ki=ram_data[DARWIN_HEAD_TILT_I_L]+(ram_data[DARWIN_HEAD_TILT_I_H]<<8);
  head_tracking_tilt.kd=ram_data[DARWIN_HEAD_TILT_D_L]+(ram_data[DARWIN_HEAD_TILT_D_H]<<8);
  head_tracking_tilt.prev_error=0;
  head_tracking_tilt.integral_part=0;
  head_tracking_tilt.integral_clamp=(ram_data[DARWIN_HEAD_TILT_I_CLAMP_L]+(ram_data[DARWIN_HEAD_TILT_I_CLAMP_H]<<8))<<9;
  head_tracking_tilt.min_angle=manager_get_cw_angle_limit(L_TILT)<<9;
  ram_data[DARWIN_HEAD_MIN_TILT_L]=manager_get_cw_angle_limit(L_TILT)%256;
  ram_data[DARWIN_HEAD_MIN_TILT_H]=manager_get_cw_angle_limit(L_TILT)/256;
  head_tracking_tilt.max_angle=manager_get_ccw_angle_limit(L_TILT)<<9;
  ram_data[DARWIN_HEAD_MAX_TILT_L]=manager_get_ccw_angle_limit(L_TILT)%256;
  ram_data[DARWIN_HEAD_MAX_TILT_H]=manager_get_ccw_angle_limit(L_TILT)/256;
  head_tracking_tilt.target_angle=0;
  head_tracking_period=(period_us<<16)/1000000;
  head_tracking_period_us=period_us;
  head_tracking_enabled=0x00;
}

void head_tracking_set_period(uint16_t period_us)
{
  head_tracking_period=(period_us<<16)/1000000;
  head_tracking_period_us=period_us;
}

uint16_t head_tracking_get_period(void)
{
  return head_tracking_period_us;
}

void head_tracking_start(void)
{
  head_tracking_enabled=0x01;
  ram_data[DARWIN_HEAD_CNTRL]|=HEAD_STATUS;
}

void head_tracking_stop(void)
{
  head_tracking_enabled=0x00;
  ram_data[DARWIN_HEAD_CNTRL]&=(~HEAD_STATUS);
}

uint8_t head_is_tracking(void)
{
  return head_tracking_enabled;
}

void head_tracking_set_pan_range(int16_t max_angle,int16_t min_angle)
{
  head_tracking_pan.max_angle=max_angle<<9;
  ram_data[DARWIN_HEAD_MAX_PAN_L]=max_angle%256;
  ram_data[DARWIN_HEAD_MAX_PAN_H]=max_angle/256;
  head_tracking_pan.min_angle=min_angle<<9;
  ram_data[DARWIN_HEAD_MIN_PAN_L]=min_angle%256;
  ram_data[DARWIN_HEAD_MIN_PAN_H]=min_angle/256;
}

void head_tracking_get_pan_range(int16_t *max_angle,int16_t *min_angle)
{
  *max_angle=head_tracking_pan.max_angle>>9;
  *min_angle=head_tracking_pan.min_angle>>9;
}

void head_tracking_set_pan_pid(uint16_t kp,uint16_t ki,uint16_t kd, uint16_t i_clamp)
{
  head_tracking_pan.kp=kp;
  ram_data[DARWIN_HEAD_PAN_P_L]=kp%256;
  ram_data[DARWIN_HEAD_PAN_P_H]=kp/256;
  head_tracking_pan.ki=ki;
  ram_data[DARWIN_HEAD_PAN_I_L]=ki%256;
  ram_data[DARWIN_HEAD_PAN_I_H]=ki/256;
  head_tracking_pan.kd=kd;
  ram_data[DARWIN_HEAD_PAN_D_L]=kd%256;
  ram_data[DARWIN_HEAD_PAN_D_H]=kd/256;
  head_tracking_pan.integral_clamp=i_clamp<<9;
  ram_data[DARWIN_HEAD_PAN_I_CLAMP_L]=i_clamp%256;
  ram_data[DARWIN_HEAD_PAN_I_CLAMP_H]=i_clamp/256;
}

void head_tracking_get_pan_pid(uint16_t *kp,uint16_t *ki,uint16_t *kd, uint16_t *i_clamp)
{
  *kp=head_tracking_pan.kp;
  *ki=head_tracking_pan.ki;
  *kd=head_tracking_pan.kd;
  *i_clamp=head_tracking_pan.integral_clamp>>9;
}

void head_tracking_set_tilt_range(int16_t max_angle,int16_t min_angle)
{
  head_tracking_tilt.max_angle=max_angle<<9;
  ram_data[DARWIN_HEAD_MAX_TILT_L]=max_angle%256;
  ram_data[DARWIN_HEAD_MAX_TILT_H]=max_angle/256;
  head_tracking_tilt.min_angle=min_angle<<9;
  ram_data[DARWIN_HEAD_MIN_TILT_L]=min_angle%256;
  ram_data[DARWIN_HEAD_MIN_TILT_H]=min_angle/256;
}

void head_tracking_get_tilt_range(int16_t *max_angle,int16_t *min_angle)
{
  *max_angle=head_tracking_tilt.max_angle>>9;
  *min_angle=head_tracking_tilt.min_angle>>9;
}

void head_tracking_set_tilt_pid(uint16_t kp,uint16_t ki,uint16_t kd, uint16_t i_clamp)
{
  head_tracking_tilt.kp=kp;
  ram_data[DARWIN_HEAD_TILT_P_L]=kp%256;
  ram_data[DARWIN_HEAD_TILT_P_H]=kp/256;
  head_tracking_tilt.ki=ki;
  ram_data[DARWIN_HEAD_TILT_I_L]=ki%256;
  ram_data[DARWIN_HEAD_TILT_I_H]=ki/256;
  head_tracking_tilt.kd=kd;
  ram_data[DARWIN_HEAD_TILT_D_L]=kd%256;
  ram_data[DARWIN_HEAD_TILT_D_H]=kd/256;
  head_tracking_tilt.integral_clamp=i_clamp<<9;
  ram_data[DARWIN_HEAD_TILT_I_CLAMP_L]=i_clamp%256;
  ram_data[DARWIN_HEAD_TILT_I_CLAMP_H]=i_clamp/256;
}

void head_tracking_get_tilt_pid(uint16_t *kp,uint16_t *ki,uint16_t *kd, uint16_t *i_clamp)
{
  *kp=head_tracking_tilt.kp;
  *ki=head_tracking_tilt.ki;
  *kd=head_tracking_tilt.kd;
  *i_clamp=head_tracking_tilt.integral_clamp>>9;
}

void head_tracking_set_target_pan(int16_t target)
{
  head_tracking_pan.target_angle=target<<9;
  ram_data[DARWIN_HEAD_PAN_TARGET_L]=target%256;
  ram_data[DARWIN_HEAD_PAN_TARGET_H]=target/256;
}

void head_tracking_set_target_tilt(int16_t target)
{
  head_tracking_tilt.target_angle=target<<9;
  ram_data[DARWIN_HEAD_TILT_TARGET_L]=target%256;
  ram_data[DARWIN_HEAD_TILT_TARGET_H]=target/256;
}

// operation functions
uint8_t head_tracking_in_range(unsigned short int address, unsigned short int length)
{
  if(ram_in_window(HEAD_BASE_ADDRESS,HEAD_MEM_LENGTH,address,length) ||
     ram_in_window(HEAD_EEPROM_ADDRESS,HEAD_EEPROM_LENGTH,address,length))
    return 0x01;
  else
    return 0x00;
}

void head_tracking_process_read_cmd(unsigned short int address,unsigned short int length,unsigned char *data)
{

}

void head_tracking_process_write_cmd(unsigned short int address,unsigned short int length,unsigned char *data)
{
  uint16_t kp,kd,ki,i_clamp;
  int16_t max,min,target;

  // head tracking control
  if(ram_in_range(DARWIN_HEAD_CNTRL,address,length))
  {
    if(data[DARWIN_HEAD_CNTRL-address]&HEAD_START)
      head_tracking_start();
    if(data[DARWIN_HEAD_CNTRL-address]&HEAD_STOP)
      head_tracking_stop();
  }
  if(ram_in_window(DARWIN_HEAD_PAN_P_L,8,address,length))
  {
    head_tracking_get_pan_pid(&kp,&kd,&ki,&i_clamp);
    if(ram_in_range(DARWIN_HEAD_PAN_P_L,address,length)) kp=(kp&0xFF00)|data[DARWIN_HEAD_PAN_P_L-address];
    if(ram_in_range(DARWIN_HEAD_PAN_P_H,address,length)) kp=(kp&0x00FF)|(data[DARWIN_HEAD_PAN_P_H-address]<<8);
    if(ram_in_range(DARWIN_HEAD_PAN_I_L,address,length)) ki=(ki&0xFF00)|data[DARWIN_HEAD_PAN_I_L-address];
    if(ram_in_range(DARWIN_HEAD_PAN_I_H,address,length)) ki=(ki&0x00FF)|(data[DARWIN_HEAD_PAN_I_H-address]<<8);
    if(ram_in_range(DARWIN_HEAD_PAN_D_L,address,length)) kd=(kd&0xFF00)|data[DARWIN_HEAD_PAN_D_L-address];
    if(ram_in_range(DARWIN_HEAD_PAN_D_H,address,length)) kd=(kd&0x00FF)|(data[DARWIN_HEAD_PAN_D_H-address]<<8);
    if(ram_in_range(DARWIN_HEAD_PAN_I_CLAMP_L,address,length)) i_clamp=(i_clamp&0xFF00)|data[DARWIN_HEAD_PAN_I_CLAMP_L-address];
    if(ram_in_range(DARWIN_HEAD_PAN_I_CLAMP_H,address,length)) i_clamp=(i_clamp&0x00FF)|(data[DARWIN_HEAD_PAN_I_CLAMP_H-address]<<8);
    head_tracking_set_pan_pid(kp,kd,ki,i_clamp);
  }
  if(ram_in_window(DARWIN_HEAD_TILT_P_L,8,address,length))
  {
    head_tracking_get_tilt_pid(&kp,&kd,&ki,&i_clamp);
    if(ram_in_range(DARWIN_HEAD_TILT_P_L,address,length)) kp=(kp&0xFF00)|data[DARWIN_HEAD_TILT_P_L-address];
    if(ram_in_range(DARWIN_HEAD_TILT_P_H,address,length)) kp=(kp&0x00FF)|(data[DARWIN_HEAD_TILT_P_H-address]<<8);
    if(ram_in_range(DARWIN_HEAD_TILT_I_L,address,length)) ki=(ki&0xFF00)|data[DARWIN_HEAD_TILT_I_L-address];
    if(ram_in_range(DARWIN_HEAD_TILT_I_H,address,length)) ki=(ki&0x00FF)|(data[DARWIN_HEAD_TILT_I_H-address]<<8);
    if(ram_in_range(DARWIN_HEAD_TILT_D_L,address,length)) kd=(kd&0xFF00)|data[DARWIN_HEAD_TILT_D_L-address];
    if(ram_in_range(DARWIN_HEAD_TILT_D_H,address,length)) kd=(kd&0x00FF)|(data[DARWIN_HEAD_TILT_D_H-address]<<8);
    if(ram_in_range(DARWIN_HEAD_TILT_I_CLAMP_L,address,length)) i_clamp=(i_clamp&0xFF00)|data[DARWIN_HEAD_TILT_I_CLAMP_L-address];
    if(ram_in_range(DARWIN_HEAD_TILT_I_CLAMP_H,address,length)) i_clamp=(i_clamp&0x00FF)|(data[DARWIN_HEAD_TILT_I_CLAMP_H-address]<<8);
    head_tracking_set_tilt_pid(kp,kd,ki,i_clamp);
  }
  if(ram_in_window(DARWIN_HEAD_MAX_PAN_L,4,address,length))
  {
    head_tracking_get_pan_range(&max,&min);
    if(ram_in_range(DARWIN_HEAD_MAX_PAN_L,address,length)) max=(max&0xFF00)|data[DARWIN_HEAD_MAX_PAN_L-address];
    if(ram_in_range(DARWIN_HEAD_MAX_PAN_H,address,length)) max=(max&0x00FF)|(data[DARWIN_HEAD_MAX_PAN_H-address]<<8);
    if(ram_in_range(DARWIN_HEAD_MIN_PAN_L,address,length)) min=(min&0xFF00)|data[DARWIN_HEAD_MIN_PAN_L-address];
    if(ram_in_range(DARWIN_HEAD_MIN_PAN_H,address,length)) min=(min&0x00FF)|(data[DARWIN_HEAD_MIN_PAN_H-address]<<8);
    head_tracking_set_pan_range(max,min);
  }
  if(ram_in_window(DARWIN_HEAD_MAX_TILT_L,4,address,length))
  {
    head_tracking_get_tilt_range(&max,&min);
    if(ram_in_range(DARWIN_HEAD_MAX_TILT_L,address,length)) max=(max&0xFF00)|data[DARWIN_HEAD_MAX_TILT_L-address];
    if(ram_in_range(DARWIN_HEAD_MAX_TILT_H,address,length)) max=(max&0x00FF)|(data[DARWIN_HEAD_MAX_TILT_H-address]<<8);
    if(ram_in_range(DARWIN_HEAD_MIN_TILT_L,address,length)) min=(min&0xFF00)|data[DARWIN_HEAD_MIN_TILT_L-address];
    if(ram_in_range(DARWIN_HEAD_MIN_TILT_H,address,length)) min=(min&0x00FF)|(data[DARWIN_HEAD_MIN_TILT_H-address]<<8);
    head_tracking_set_tilt_range(max,min);
  }
  if(ram_in_window(DARWIN_HEAD_PAN_TARGET_L,2,address,length))
  {
    target=(head_tracking_pan.target_angle>>9);
    if(ram_in_range(DARWIN_HEAD_PAN_TARGET_L,address,length))
    { 
      target=(target&0xFF00)|data[DARWIN_HEAD_PAN_TARGET_L-address];
      ram_data[DARWIN_HEAD_PAN_TARGET_L]=data[DARWIN_HEAD_PAN_TARGET_L-address];
    }
    if(ram_in_range(DARWIN_HEAD_PAN_TARGET_H,address,length))
    {
      target=(target&0x00FF)|(data[DARWIN_HEAD_PAN_TARGET_H-address]<<8);
      ram_data[DARWIN_HEAD_PAN_TARGET_H]=data[DARWIN_HEAD_PAN_TARGET_H-address];
    }
    head_tracking_pan.target_angle=(target<<9);
  }
  if(ram_in_window(DARWIN_HEAD_TILT_TARGET_L,2,address,length))
  {
    target=(head_tracking_tilt.target_angle>>9);
    if(ram_in_range(DARWIN_HEAD_TILT_TARGET_L,address,length))
    {
      target=(target&0xFF00)|data[DARWIN_HEAD_TILT_TARGET_L-address];
      ram_data[DARWIN_HEAD_TILT_TARGET_L]=data[DARWIN_HEAD_TILT_TARGET_L-address];
    }
    if(ram_in_range(DARWIN_HEAD_TILT_TARGET_H,address,length))
    {
      target=(target&0x00FF)|(data[DARWIN_HEAD_TILT_TARGET_H-address]<<8);
      ram_data[DARWIN_HEAD_TILT_TARGET_H]=data[DARWIN_HEAD_TILT_TARGET_H-address];
    }
    head_tracking_tilt.target_angle=(target<<9);
  }
}

// motion manager process function
void head_tracking_process(void)
{
  int64_t pan_error, tilt_error;
  int64_t derivative_pan=0, derivative_tilt=0;

  if(head_tracking_enabled)
  {
    ram_data[DARWIN_HEAD_CNTRL]|=HEAD_STATUS;
    if(manager_get_module(L_PAN)==MM_HEAD)
    {
      pan_error = head_tracking_pan.target_angle-manager_current_angles[L_PAN];
      head_tracking_pan.integral_part+=(pan_error*head_tracking_period)>>16;
      if(head_tracking_pan.integral_part>head_tracking_pan.integral_clamp)
        head_tracking_pan.integral_part=head_tracking_pan.integral_clamp;
      else if(head_tracking_pan.integral_part<-head_tracking_pan.integral_clamp)
        head_tracking_pan.integral_part=-head_tracking_pan.integral_clamp;
      derivative_pan = ((pan_error - head_tracking_pan.prev_error)<<16)/head_tracking_period;
      manager_current_angles[L_PAN]+=((head_tracking_pan.kp*pan_error)>>16);
      manager_current_angles[L_PAN]+=((head_tracking_pan.ki*head_tracking_pan.integral_part)>>16);
      manager_current_angles[L_PAN]+=((head_tracking_pan.kd*derivative_pan)>>16);
      if(manager_current_angles[L_PAN]>head_tracking_pan.max_angle)
        manager_current_angles[L_PAN]=head_tracking_pan.max_angle;
      else if(manager_current_angles[L_PAN]<head_tracking_pan.min_angle)
        manager_current_angles[L_PAN]=head_tracking_pan.min_angle;
      head_tracking_pan.prev_error = pan_error;
    }
    if(manager_get_module(L_TILT)==MM_HEAD)
    { 
      tilt_error = head_tracking_tilt.target_angle-manager_current_angles[L_TILT];
      head_tracking_tilt.integral_part+=(tilt_error*head_tracking_period)>>16;
      if(head_tracking_tilt.integral_part>head_tracking_tilt.integral_clamp)
        head_tracking_tilt.integral_part=head_tracking_tilt.integral_clamp;
      else if(head_tracking_tilt.integral_part<-head_tracking_tilt.integral_clamp)
        head_tracking_tilt.integral_part=-head_tracking_tilt.integral_clamp;
      derivative_tilt = ((tilt_error - head_tracking_tilt.prev_error)<<16)/head_tracking_period;
      manager_current_angles[L_TILT]+=((head_tracking_tilt.kp*tilt_error)>>16);
      manager_current_angles[L_TILT]+=((head_tracking_tilt.ki*head_tracking_tilt.integral_part)>>16);
      manager_current_angles[L_TILT]+=((head_tracking_tilt.kd*derivative_tilt)>>16);
      if(manager_current_angles[L_TILT]>head_tracking_tilt.max_angle)
        manager_current_angles[L_TILT]=head_tracking_tilt.max_angle;
      else if(manager_current_angles[L_TILT]<head_tracking_tilt.min_angle)
        manager_current_angles[L_TILT]=head_tracking_tilt.min_angle;
      head_tracking_tilt.prev_error = tilt_error;
    }
  }
  else
    ram_data[DARWIN_HEAD_CNTRL]&=(~HEAD_STATUS);
}

