#include "stm32f1xx_hal.h"
#include "stm32_time.h"
#include "comm.h"
#include "usart3.h"
#include "dynamixel_slave.h"

#define ENABLE_LED_RX_GPIO_CLK      __HAL_RCC_GPIOC_CLK_ENABLE()
#define LED_RX_PIN                  GPIO_PIN_14                
#define LED_RX_GPIO_PORT            GPIOC                       

TIM_HandleTypeDef us_timer_Handle;
unsigned long long int counts;

void SysTick_Handler(void)
{
  HAL_IncTick();
}

void TIM7_IRQHandler(void)
{
  if(__HAL_TIM_GET_FLAG(&us_timer_Handle, TIM_FLAG_UPDATE) != RESET)
  {
    if(__HAL_TIM_GET_IT_SOURCE(&us_timer_Handle, TIM_IT_UPDATE) !=RESET)
    {
      __HAL_TIM_CLEAR_IT(&us_timer_Handle, TIM_IT_UPDATE);
      counts+=1;
    }
  }
}

void timer_init(void)
{
  us_timer_Handle.Instance=TIM7;
  us_timer_Handle.Init.Period=0xFFFF;
  us_timer_Handle.Init.Prescaler=0;
  us_timer_Handle.Init.ClockDivision=TIM_CLOCKDIVISION_DIV1;
  us_timer_Handle.Init.CounterMode=TIM_COUNTERMODE_UP;
  us_timer_Handle.Init.RepetitionCounter=0;
  HAL_TIM_Base_Init(&us_timer_Handle);
  __HAL_RCC_TIM7_CLK_ENABLE();
  __HAL_TIM_SET_COUNTER(&us_timer_Handle,0);

  // enable interrupts
  HAL_NVIC_SetPriority(TIM7_IRQn, 3, 3);
  HAL_NVIC_EnableIRQ(TIM7_IRQn);
  HAL_TIM_Base_Start_IT(&us_timer_Handle);

  counts=0;
}

unsigned long long int timer_get_counts(void)
{
  return (counts<<16)+__HAL_TIM_GET_COUNTER(&us_timer_Handle);
}

void gpio_init(void)
{
  GPIO_InitTypeDef GPIO_InitStructure;

  ENABLE_LED_RX_GPIO_CLK;

  GPIO_InitStructure.Pin = LED_RX_PIN;
  GPIO_InitStructure.Mode = GPIO_MODE_OUTPUT_PP;
  GPIO_InitStructure.Speed = GPIO_SPEED_HIGH;
  GPIO_InitStructure.Pull = GPIO_NOPULL;
  HAL_GPIO_Init(LED_RX_GPIO_PORT, &GPIO_InitStructure);

  HAL_GPIO_WritePin(LED_RX_GPIO_PORT,LED_RX_PIN,GPIO_PIN_SET);
}

void on_ping(void)
{
  HAL_GPIO_TogglePin(GPIOC,GPIO_PIN_14);
}

int main(void)
{
  TUSART_IRQ_Priorities priorities;
  TDynamixelSlave dyn_slave;
  uint32_t clocks_per_us;
  UART_InitTypeDef Init;
  TTime time;
  TComm comm;

  /* initialize the HAL module */
  HAL_Init();
  /* initialize the timer */
  timer_init();
  clocks_per_us=(SystemCoreClock/1000000)+1;
  time_init(&time,clocks_per_us,timer_get_counts);
  /* initialize the GPIO */
  gpio_init();
  /* initialize the comm object */
  comm_init(&comm,0x01,&time);
  Init.BaudRate     = 57600;
  Init.WordLength   = UART_WORDLENGTH_8B;
  Init.StopBits     = UART_STOPBITS_1;
  Init.Parity       = UART_PARITY_NONE;
  Init.Mode         = UART_MODE_TX_RX;
  Init.HwFlowCtl    = UART_HWCONTROL_NONE;
  Init.OverSampling = UART_OVERSAMPLING_16;
  priorities.irq_priority=0;
  priorities.irq_subpriority=0;
  priorities.dma_rx_priority=1;
  priorities.dma_rx_subpriority=1;
  priorities.dma_tx_priority=1;
  priorities.dma_tx_subpriority=0;
  usart3_init(&comm,&Init,&priorities);
  dyn_slave_init(&dyn_slave,&comm,0xC0);
  dyn_slave.on_ping=on_ping;
  dyn_slave_set_return_delay(&dyn_slave,250);

  while(1)                             /* main function does not return */
  {
    dyn_slave_loop(&dyn_slave);
  }
}

