#ifndef _MOTION_PATTERNS_H
#define _MOTION_PATTERNS_H

#include "patterns.h"
#include "frame_buffer_control.h"

#ifndef MTN_X_LED_SPACING
  #error "Please, specify the LED spacing in mm in the X direction with the MTN_X_LED_SPACING macro"
#endif

#ifndef MTN_Y_LED_SPACING
  #error "Please, specify the LED spacing in mm in the Y direction with the MTN_Y_LED_SPACING macro"
#endif

#define MOTION_GROUP                0x30

typedef enum {MTN_RAMP=0x01,MTN_TRIANGLE=0x02,MTN_CIRCLE=0x03,MTN_NONE=0x04} mtn_id_t;

typedef struct
{
  short int max_coord_x;
  short int max_coord_y;
  short int min_coord_x;
  short int min_coord_y;
  unsigned char direction;
}TMTNRampData;

typedef struct
{
  short int max_coord_x;
  short int max_coord_y;
  short int min_coord_x;
  short int min_coord_y;
  unsigned char direction;
}TMTNTriangleData;

typedef struct
{
  short int radius_mm;
  short int center_x;
  short int center_y;
  short int dummy1;
  unsigned char dummy2;
}TMTNCircleData;

typedef union
{
  TMTNRampData ramp_data;
  TMTNTriangleData triangle_data;
  TMTNCircleData circle_data;
}TMTNData;

#pragma pack (push, 1)
typedef struct
{
  mtn_id_t motion_id;
  unsigned short int speed_mm_s;
  TMTNData motion_data;
  // internal data
  short int current_col;
  short int current_row;
  unsigned short int current_time;
}TMotion;
#pragma pack (pop)

#define MTN_ID_OFFSET                0
#define MTN_SPEED_OFFSET             1
#define MTN_DATA                     3
#define MTN_CURRENT_COL_OFFSET       12
#define MTN_CURRENT_ROW_OFFSET       14
#define MTN_CURRENT_TIME_OFFSET      16
void mtn_general(TLEDArea *area,TMotion *pattern_data,unsigned char buffer_id,unsigned short int period,TFBControl *control);

void mtn_compute(TMotion *pattern_data);

#endif
