#include "motion_patterns.h"
#include <math.h>

// private functions
void mtn_ramp(TMotion *pattern_data)
{
  unsigned short int ms_pixel_x,delta_x_mm;
  unsigned short int ms_pixel_y,delta_y_mm;
  unsigned short int dist_mm;

  delta_x_mm=(pattern_data->motion_data.ramp_data.max_coord_x-pattern_data->motion_data.ramp_data.min_coord_x)*MTN_X_LED_SPACING;
  delta_y_mm=(pattern_data->motion_data.ramp_data.max_coord_y-pattern_data->motion_data.ramp_data.min_coord_y)*MTN_Y_LED_SPACING;

  dist_mm=sqrt(pow(delta_x_mm,2)+pow(delta_y_mm,2));

  ms_pixel_x=(dist_mm*1000)/(delta_x_mm*pattern_data->speed_mm_s);
  ms_pixel_y=(dist_mm*1000)/(delta_y_mm*pattern_data->speed_mm_s);

  if(pattern_data->motion_data.ramp_data.direction==0x01)
  {
    pattern_data->current_col=pattern_data->motion_data.ramp_data.min_coord_x+(pattern_data->current_time/ms_pixel_x);
    pattern_data->current_row=pattern_data->motion_data.ramp_data.min_coord_y+(pattern_data->current_time/ms_pixel_y);

    if((pattern_data->current_col>=pattern_data->motion_data.ramp_data.max_coord_x) &&
       (pattern_data->current_row>=pattern_data->motion_data.ramp_data.max_coord_y))
    {
      pattern_data->current_col=pattern_data->motion_data.ramp_data.min_coord_x;
      pattern_data->current_row=pattern_data->motion_data.ramp_data.min_coord_y;
      pattern_data->current_time=0;
    }
  }
  else
  {
    pattern_data->current_col=pattern_data->motion_data.ramp_data.max_coord_x-(pattern_data->current_time/ms_pixel_x);
    pattern_data->current_row=pattern_data->motion_data.ramp_data.max_coord_y-(pattern_data->current_time/ms_pixel_y);

    if((pattern_data->current_col<=pattern_data->motion_data.ramp_data.min_coord_x) &&
       (pattern_data->current_row<=pattern_data->motion_data.ramp_data.min_coord_y))
    {
      pattern_data->current_col=pattern_data->motion_data.ramp_data.max_coord_x;
      pattern_data->current_row=pattern_data->motion_data.ramp_data.max_coord_y;
      pattern_data->current_time=0;
    }
  }
}

void mtn_triangle(TMotion *pattern_data)
{
  unsigned short int ms_pixel_x,delta_x_mm;
  unsigned short int ms_pixel_y,delta_y_mm;
  unsigned short int dist_mm;

  delta_x_mm=(pattern_data->motion_data.ramp_data.max_coord_x-pattern_data->motion_data.ramp_data.min_coord_x)*MTN_X_LED_SPACING;
  delta_y_mm=(pattern_data->motion_data.ramp_data.max_coord_y-pattern_data->motion_data.ramp_data.min_coord_y)*MTN_Y_LED_SPACING;

  dist_mm=sqrt(pow(delta_x_mm,2)+pow(delta_y_mm,2));

  ms_pixel_x=(dist_mm*1000)/(delta_x_mm*pattern_data->speed_mm_s);
  ms_pixel_y=(dist_mm*1000)/(delta_y_mm*pattern_data->speed_mm_s);

  if(pattern_data->motion_data.ramp_data.direction==0x01)
  {
    pattern_data->current_col=pattern_data->motion_data.ramp_data.min_coord_x+(pattern_data->current_time/ms_pixel_x);
    pattern_data->current_row=pattern_data->motion_data.ramp_data.min_coord_y+(pattern_data->current_time/ms_pixel_y);

    if((pattern_data->current_col>=pattern_data->motion_data.ramp_data.max_coord_x) &&
       (pattern_data->current_row>=pattern_data->motion_data.ramp_data.max_coord_y))
    {
      pattern_data->current_time=0;
      pattern_data->motion_data.ramp_data.direction=0x00;
    }
  }
  else
  {
    pattern_data->current_col=pattern_data->motion_data.ramp_data.max_coord_x-(pattern_data->current_time/ms_pixel_x);
    pattern_data->current_row=pattern_data->motion_data.ramp_data.max_coord_y-(pattern_data->current_time/ms_pixel_y);

    if((pattern_data->current_col<=pattern_data->motion_data.ramp_data.min_coord_x) &&
       (pattern_data->current_row<=pattern_data->motion_data.ramp_data.min_coord_y))
    {
      pattern_data->current_time=0;
      pattern_data->motion_data.ramp_data.direction=0x01;
    }
  }
}

// public functions
void mtn_general(TLEDArea *area,TMotion *pattern_data,unsigned char buffer_id,unsigned short int period,TFBControl *control)
{
  pattern_data->current_time+=period;
}

void mtn_compute(TMotion *pattern_data)
{
  // image pattern motion
  switch(pattern_data->motion_id)
  {
    case MTN_RAMP: mtn_ramp(pattern_data);
                   break;
    case MTN_TRIANGLE: mtn_triangle(pattern_data);
                       break;
    case MTN_CIRCLE:
                     break;
    case MTN_NONE: pattern_data->current_col=0;
                   pattern_data->current_row=0;
                   break;
  }
}
